"""
    Copyright 2018 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

import os
import re
from abc import ABC, abstractmethod
from collections.abc import Mapping, Sequence
from typing import Generic, Optional, TypeVar

from jinja2 import Environment, PackageLoader

from inmanta.ast import CompilerException, ModifiedAfterFreezeException
from inmanta.ast.statements import AssignStatement
from inmanta.ast.statements.generator import Constructor, IndexCollisionException
from inmanta.execute.runtime import OptionVariable
from inmanta.module import ModuleV2InV1PathException


def bold(content: Optional[str] = None) -> str:
    if content is None:
        return "\033[1m"
    return f"\033[1m{content}\033[0m"


def underline(content: Optional[str] = None) -> str:
    if content is None:
        return "\033[4m"
    return f"\033[4m{content}\033[0m"


def noformat(content: Optional[str] = None) -> str:
    return "\033[0m"


CUSTOM_FILTERS = {"bold": bold, "underline": underline, "noformat": noformat}


class ExplainerABC(ABC):
    """
    Abstract base class for explainers. This class is purposely kept non-Generic to present a public interface that is invariant
    of the compiler exception type. This allows correct typing of sequences of explainers.
    """

    @abstractmethod
    def explain(self, problem: CompilerException) -> list[str]:
        pass


Explainable = TypeVar("Explainable", bound=CompilerException)


class Explainer(Generic[Explainable], ExplainerABC, ABC):
    """
    Abstract explainer, Generic in the compiler exception subtype to allow correct typing of the exception for subtype-specific
    explanation logic.
    Concrete subclasses must not be generic in the exception type because this would break explainable checking.
    """

    explainable_type: type[Explainable]

    def explain(self, problem: CompilerException) -> list[str]:
        """
        Returns a list of explanations for this exception. If neither the exception or any of its causes (recursively)
        is explainable by this explainer, returns an empty list.
        """
        allcauses: set[CompilerException] = set()
        work: list[CompilerException] = [problem]
        while work:
            w = work.pop()
            allcauses.add(w)
            work.extend(w.get_causes())

        return [self.do_explain(c) for c in allcauses if isinstance(c, self.explainable_type)]

    @abstractmethod
    def do_explain(self, problem: Explainable) -> str:
        """
        Explain a single exception, explainable by this explainer. Does not recurse on its causes.
        """
        ...


class JinjaExplainer(Explainer[Explainable], ABC):
    """
    Abstract explainer for explanations based on a Jinja template.

    :param template: path to the Jinja template to use for the explanation.
    """

    def __init__(self, template: str) -> None:
        self.template: str = template

    def get_template(self, problem: Explainable) -> str:
        path = os.path.join(os.path.dirname(__file__), self.template)
        with open(path, encoding="utf-8") as fh:
            return fh.read()

    def do_explain(self, problem: Explainable) -> str:
        env = Environment(loader=PackageLoader("inmanta.compiler.help"))
        for name, filter in CUSTOM_FILTERS.items():
            env.filters[name] = filter

        template = env.get_template(self.template)
        return template.render(**self.get_arguments(problem))

    @abstractmethod
    def get_arguments(self, problem: Explainable) -> Mapping[str, object]:
        """
        Returns a mapping for names that are used in the Jinja template.
        """
        ...


class ModifiedAfterFreezeExplainer(JinjaExplainer[ModifiedAfterFreezeException]):
    """
    Explainer for ModifiedAfterFreezeException.
    """

    explainable_type: type[ModifiedAfterFreezeException] = ModifiedAfterFreezeException

    def __init__(self) -> None:
        super().__init__("modified_after_freeze.j2")

    def build_reverse_hint(self, problem: ModifiedAfterFreezeException) -> str:
        if isinstance(problem.stmt, AssignStatement):
            return "{}.{} = {}".format(
                problem.stmt.rhs.pretty_print(),
                problem.attribute.get_name(),
                problem.stmt.lhs.pretty_print(),
            )

        if isinstance(problem.stmt, Constructor):
            # find right parameter:
            attr = problem.attribute.end.get_name()
            if attr not in problem.stmt.get_attributes():
                attr_rhs = "?"
            else:
                attr_rhs = problem.stmt.get_attributes()[attr].pretty_print()
            return f"{attr_rhs}.{problem.attribute.get_name()} = {problem.stmt.pretty_print()}"

    def get_arguments(self, problem: ModifiedAfterFreezeException) -> Mapping[str, object]:
        return {
            "relation": problem.attribute.get_name(),
            "instance": problem.instance,
            "values": problem.resultvariable.value,
            "value": problem.value,
            "location": problem.location,
            "reverse": problem.reverse,
            "reverse_example": "" if not problem.reverse else self.build_reverse_hint(problem),
            "optional": isinstance(problem.resultvariable, OptionVariable),
        }


class ModuleV2InV1PathExplainer(JinjaExplainer[ModuleV2InV1PathException]):
    """
    Explainer for ModuleV2InV1PathException
    """

    explainable_type: type[ModuleV2InV1PathException] = ModuleV2InV1PathException

    def __init__(self) -> None:
        super().__init__("module_v2_in_v1_path.j2")

    def get_arguments(self, problem: ModuleV2InV1PathException) -> Mapping[str, object]:
        v2_source_configured: bool = problem.project.module_v2_source_configured() if problem.project is not None else False
        return {
            "name": problem.module.name,
            "path": problem.module.path,
            "project": problem.project is not None,
            "v2_source_configured": v2_source_configured,
        }


class IndexCollisionExplainer(JinjaExplainer[IndexCollisionException]):
    """
    Explainer for IndexCollisionException
    """

    explainable_type: type[IndexCollisionException] = IndexCollisionException

    def __init__(self) -> None:
        super().__init__("index_collision.j2")

    def get_arguments(self, problem: IndexCollisionException) -> Mapping[str, object]:
        return {
            "constructor_str": problem.constructor.pretty_print(),
            "constructor_loc": problem.constructor.location,
            "constructor_name": problem.constructor.class_type,
            "collisions": {",".join(index): instance for index, instance in problem.collisions.items()},
        }


def escape_ansi(line: str) -> str:
    ansi_escape = re.compile(r"(\x9B|\x1B\[)[0-?]*[ -/]*[@-~]")
    return ansi_escape.sub("", line)


class ExplainerFactory:
    def get_explainers(self) -> Sequence[ExplainerABC]:
        return [ModifiedAfterFreezeExplainer(), ModuleV2InV1PathExplainer(), IndexCollisionExplainer()]

    def explain(self, problem: CompilerException) -> list[str]:
        return [explanation for explainer in self.get_explainers() for explanation in explainer.explain(problem)]

    def explain_and_format(self, problem: CompilerException, plain: bool = True) -> Optional[str]:
        """
        :param plain: remove tty color codes, only return plain text
        """
        raw = self.explain(problem)
        if not raw:
            return None
        else:
            pre = """
\033[1mException explanation
=====================\033[0m
"""
            pre += "\n\n".join(raw)

            if not plain:
                return pre
            else:
                return escape_ansi(pre)



"""
    Copyright 2024 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

from asyncpg import Connection


async def update(connection: Connection) -> None:
    schema = """
    -- Add the 'is_marked_for_deletion' column to mark an environment as in the process of being deleted.

    ALTER TABLE public.environment ADD COLUMN is_marked_for_deletion BOOLEAN DEFAULT FALSE;

    """
    await connection.execute(schema)



"""
    Copyright 2017 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

import asyncio
import collections.abc
import inspect
import os
import subprocess
import warnings
from collections import abc
from typing import TYPE_CHECKING, Any, Callable, Literal, Mapping, Optional, Sequence, Type, TypeVar

import inmanta.ast.type as inmanta_type
from inmanta import const, protocol, util
from inmanta.ast import LocatableString, Location, Namespace, Range, RuntimeException, TypeNotFoundException, WithComment
from inmanta.ast.type import NamedType
from inmanta.config import Config
from inmanta.execute.proxy import DynamicProxy
from inmanta.execute.runtime import QueueScheduler, Resolver, ResultVariable
from inmanta.execute.util import NoneValue, Unknown
from inmanta.stable_api import stable_api
from inmanta.warnings import InmantaWarning

T = TypeVar("T")

if TYPE_CHECKING:
    from inmanta.ast.statements import DynamicStatement
    from inmanta.ast.statements.call import FunctionCall
    from inmanta.compiler import Compiler


class PluginDeprecationWarning(InmantaWarning):
    pass


@stable_api
class Context:
    """
    An instance of this class is used to pass context to the plugin
    """

    __client: Optional["protocol.Client"] = None
    __sync_client = None

    @classmethod
    def __get_client(cls) -> "protocol.Client":
        if cls.__client is None:
            cls.__client = protocol.Client("compiler")
        return cls.__client

    def __init__(
        self, resolver: Resolver, queue: QueueScheduler, owner: "FunctionCall", plugin: "Plugin", result: ResultVariable
    ) -> None:
        self.resolver = resolver
        self.queue = queue
        self.owner = owner
        self.plugin = plugin
        self.result = result
        self.compiler = queue.get_compiler()

    def get_resolver(self) -> Resolver:
        return self.resolver

    def get_type(self, name: LocatableString) -> inmanta_type.Type:
        """
        Get a type from the configuration model.
        """
        try:
            return self.queue.get_types()[str(name)]
        except KeyError:
            raise TypeNotFoundException(name, self.owner.namespace)

    def get_queue_scheduler(self) -> QueueScheduler:
        return self.queue

    def get_environment_id(self) -> str:
        env = str(Config.get("config", "environment", None))

        if env is None:
            raise Exception("The environment of the model should be configured in config>environment")

        return env

    def get_compiler(self) -> "Compiler":
        return self.queue.get_compiler()

    def get_data_dir(self) -> str:
        """
        Get the path to the data dir (and create if it does not exist yet
        """
        data_dir = os.path.join("data", self.plugin.namespace.get_full_name())

        if not os.path.exists(data_dir):
            os.makedirs(data_dir, exist_ok=True)

        return data_dir

    def get_client(self) -> "protocol.Client":
        return self.__class__.__get_client()

    def get_sync_client(self) -> "protocol.SyncClient":
        if self.__class__.__sync_client is None:
            self.__class__.__sync_client = protocol.SyncClient("compiler")
        return self.__class__.__sync_client

    def run_sync(self, function: Callable[[], abc.Awaitable[T]], timeout: int = 5) -> T:
        """
        Execute the async function and return its result. This method uses this thread's current (not running) event loop if
        there is one, otherwise it creates a new one. The main use for this function is to use the inmanta internal rpc to
        communicate with the server.

        :param function: The async function to execute. This function should return a yieldable object.
        :param timeout: A timeout for the async function.
        :return: The result of the async call.
        :raises ConnectionRefusedError: When the function timeouts this exception is raised.
        """
        with_timeout: abc.Awaitable[T] = asyncio.wait_for(function(), timeout)
        try:
            return util.ensure_event_loop().run_until_complete(with_timeout)
        except TimeoutError:
            raise ConnectionRefusedError()


@stable_api
class PluginMeta(type):
    """
    A metaclass that keeps track of concrete plugin subclasses. This class is responsible for all plugin registration.
    """

    def __new__(cls, name: str, bases: tuple[type, ...], dct: dict[str, object]) -> type:
        subclass = type.__new__(cls, name, bases, dct)
        if hasattr(subclass, "__function_name__"):
            cls.add_function(subclass)
        return subclass

    __functions: dict[str, type["Plugin"]] = {}

    @classmethod
    def add_function(cls, plugin_class: type["Plugin"]) -> None:
        """
        Add a function plugin class
        """
        cls.__functions[plugin_class.__fq_plugin_name__] = plugin_class

    @classmethod
    def get_functions(cls) -> dict[str, "Type[Plugin]"]:
        """
        Get all functions that are registered
        """
        return dict(cls.__functions)

    @classmethod
    def clear(cls, inmanta_module: Optional[str] = None) -> None:
        """
        Clears registered plugin functions.

        :param inmanta_module: Clear plugin functions for a specific inmanta module. If omitted, clears all registered plugin
            functions.
        """
        if inmanta_module is not None:
            top_level: str = f"{const.PLUGINS_PACKAGE}.{inmanta_module}"
            cls.__functions = {
                fq_name: plugin_class
                for fq_name, plugin_class in cls.__functions.items()
                if plugin_class.__module__ != top_level and not plugin_class.__module__.startswith(f"{top_level}.")
            }
        else:
            cls.__functions = {}


class Null(inmanta_type.Type):
    """
    This custom type is used for the validation of plugins which only
    accept null as an argument or return value.
    """

    def validate(self, value: Optional[object]) -> bool:
        if isinstance(value, NoneValue):
            return True

        raise RuntimeException(None, f"Invalid value '{value}', expected {self.type_string()}")

    def type_string(self) -> str:
        return "null"

    def type_string_internal(self) -> str:
        return self.type_string()


# Define some types which are used in the context of plugins.
PLUGIN_TYPES = {
    "any": inmanta_type.Type(),  # Any value will pass validation
    "expression": inmanta_type.Type(),  # Any value will pass validation
    "null": Null(),  # Only NoneValue will pass validation
    None: Null(),  # Only NoneValue will pass validation
}


class PluginValue:
    """
    Base class for all values that go in and out of a plugin: arguments and return value.

    The class has two class attributes that should be set in the different subclasses:
    :attr VALUE_TYPE: The type of io value it is, argument or return value
    :attr VALUE_NAME: The name of the io value, the argument name or "return value"

    These attributes are only used for better error reporting.
    """

    VALUE_TYPE: str = ""
    VALUE_NAME: str = ""

    def __init__(self, type_expression: object) -> None:
        self.type_expression = type_expression
        self._resolved_type: Optional[inmanta_type.Type] = None

    @property
    def resolved_type(self) -> inmanta_type.Type:
        """
        Get the resolved type of this plugin io.  The resolved type can only be accessed
        once this object has been normalized (which happens during the plugin normalization).
        """
        if self._resolved_type is None:
            raise RuntimeException(
                stmt=None,
                msg=f"{type(self).__name__} {self.VALUE_NAME} ({repr(self.type_expression)}) has not been normalized, "
                "its resolved type can't be accessed.",
            )
        return self._resolved_type

    def resolve_type(self, plugin: "Plugin", resolver: Namespace) -> inmanta_type.Type:
        """
        Convert the string representation of this argument's type to a type.
        If no type annotation is present or if the type annotation allows any type to be passed
        as argument, then None is returned.

        :param plugin: The plugin that this argument is part of.
        :param resolver: The namespace that can be used to resolve the type annotation of this
            argument.
        """
        if self.type_expression in PLUGIN_TYPES:
            self._resolved_type = PLUGIN_TYPES[self.type_expression]
            return self._resolved_type

        if not isinstance(self.type_expression, str):
            raise RuntimeException(
                stmt=None,
                msg="Bad annotation in plugin %s for %s, expected str but got %s (%s)"
                % (plugin.get_full_name(), self.VALUE_NAME, type(self.type_expression).__name__, self.type_expression),
            )

        plugin_line: Range = Range(plugin.location.file, plugin.location.lnr, 1, plugin.location.lnr + 1, 1)
        locatable_type: LocatableString = LocatableString(self.type_expression, plugin_line, 0, resolver)
        self._resolved_type = inmanta_type.resolve_type(locatable_type, resolver)
        return self._resolved_type

    def validate(self, value: object) -> bool:
        """
        Validate that the given value can be passed to this argument.  Returns True if the value is known
        and valid, False if the value is unknown, and raises a ValueError is the value is not of the
        expected type.

        :param value: The value to validate
        """
        if isinstance(value, Unknown):
            # Value is not known, it can not be validated
            return False

        # Validate the value, use custom validate method of the type if it exists
        return self.resolved_type.validate(value)


class PluginArgument(PluginValue):
    """
    Represents the argument of an Inmanta plugin.
    """

    VALUE_TYPE = "argument value"

    # Marker used to indicate that a plugin argument has no default value.
    NO_DEFAULT_VALUE_SET = object()

    def __init__(
        self,
        arg_name: str,
        arg_type: object,
        arg_position: Optional[int] = None,
        default_value: object = NO_DEFAULT_VALUE_SET,
    ) -> None:
        super().__init__(arg_type)
        self.arg_name = arg_name
        self.arg_type = self.type_expression
        self.arg_position = arg_position
        self.is_kw_only_argument = arg_position is None
        self._default_value = default_value
        self.VALUE_NAME = self.arg_name

    @property
    def default_value(self) -> Optional[object]:
        if not self.has_default_value():
            raise Exception("PluginArgument doesn't have a default value")
        return self._default_value

    def has_default_value(self) -> bool:
        """
        Return True iff this plugin argument has a default value set.
        """
        return self._default_value is not self.NO_DEFAULT_VALUE_SET

    def __str__(self) -> str:
        if self.has_default_value():
            return "%s: %s = %s" % (self.arg_name, repr(self.arg_type), str(self.default_value))
        else:
            return "%s: %s" % (self.arg_name, repr(self.arg_type))


class PluginReturn(PluginValue):
    """
    Represent the return type of an Inmanta plugin.
    """

    VALUE_TYPE = "returned value"
    VALUE_NAME = "return value"


class Plugin(NamedType, WithComment, metaclass=PluginMeta):
    """
    This class models a plugin that can be called from the language.
    """

    deprecated: bool = False
    replaced_by: Optional[str] = None

    def __init__(self, namespace: Namespace) -> None:
        self.ns = namespace
        self.namespace = namespace

        # The index of the Context attribute.
        self._context: int = -1

        # Load the signature and build all the PluginArgument objects corresponding to it
        self.args: list[PluginArgument] = list()
        self.var_args: Optional[PluginArgument] = None
        self.kwargs: dict[str, PluginArgument] = dict()
        self.var_kwargs: Optional[PluginArgument] = None
        self.all_args: dict[str, PluginArgument] = dict()
        self.return_type: PluginReturn = PluginReturn("null")
        if hasattr(self.__class__, "__function__"):
            self._load_signature(self.__class__.__function__)

        self.new_statement = None

        filename: Optional[str] = inspect.getsourcefile(self.__class__.__function__)
        assert filename is not None
        try:
            line: int = inspect.getsourcelines(self.__class__.__function__)[1] + 1
        except OSError:
            # In case of bytecompiled code there is no source line
            line = 1

        if self.__class__.__function__.__doc__:
            self.comment = self.__class__.__function__.__doc__

        self.location = Location(filename, line)

    def normalize(self) -> None:
        self.resolver = self.namespace

        # Resolve all the types that we expect to receive as input of our plugin
        for arg in self.all_args.values():
            arg.resolve_type(self, self.resolver)
        if self.var_args is not None:
            self.var_args.resolve_type(self, self.resolver)
        if self.var_kwargs is not None:
            self.var_kwargs.resolve_type(self, self.resolver)

        self.return_type.resolve_type(self, self.resolver)

    def _load_signature(self, function: Callable[..., object]) -> None:
        """
        Load the signature from the given python function, and update the relevant attributes
        of this object.
        """
        # Reset relevant object attributes
        self.args = list()
        self.var_args = None
        self.kwargs = dict()
        self.var_kwargs = None
        self.all_args = dict()

        # Inspect the function to get its arguments and annotations
        arg_spec = inspect.getfullargspec(function)

        # Load the return annotation.  If not return annotation is provided, the returned
        # type is "any".
        if "return" not in arg_spec.annotations:
            self.return_type = PluginReturn("any")
        else:
            self.return_type = PluginReturn(arg_spec.annotations["return"])

        def get_annotation(arg: str) -> object:
            """
            Get the annotation for a specific argument, and if none exists, raise an exception
            """
            if arg not in arg_spec.annotations:
                raise RuntimeException(
                    stmt=None,
                    msg=f"All arguments of plugin {repr(self.get_full_name())} should be annotated: "
                    f"{repr(arg)} has no annotation",
                )

            return arg_spec.annotations[arg]

        if arg_spec.varargs is not None:
            # We have a catch-all positional arguments
            self.var_args = PluginArgument(
                arg_name=arg_spec.varargs,
                arg_type=get_annotation(arg_spec.varargs),
            )

        if arg_spec.varkw is not None:
            # We have a catch-all keyword arguments
            self.var_kwargs = PluginArgument(
                arg_name=arg_spec.varkw,
                arg_type=get_annotation(arg_spec.varkw),
            )

        # Save all positional arguments
        defaults_start_at = len(arg_spec.args) - len(arg_spec.defaults or [])
        for position, arg in enumerate(arg_spec.args):
            annotation = get_annotation(arg)
            if annotation == Context:
                self._context = position
                continue

            # Resolve the default before changing the position because
            # of the context presence, as the defaults list definitely
            # takes into account the context presence.
            default = (
                arg_spec.defaults[position - defaults_start_at]
                if position >= defaults_start_at
                else PluginArgument.NO_DEFAULT_VALUE_SET
            )

            if self._context != -1:
                # If we have a context argument, the position index
                # needs to be adapted as this context object can never be passed
                # from the model.
                position -= 1

            argument = PluginArgument(
                arg_name=arg,
                arg_type=annotation,
                arg_position=position,
                default_value=default,
            )

            # This is a positional argument, we register it now
            self.args.append(argument)
            self.all_args[arg] = argument

        # Save all key-word arguments
        for arg in arg_spec.kwonlyargs:
            argument = PluginArgument(
                arg_name=arg,
                arg_type=get_annotation(arg),
                default_value=(
                    arg_spec.kwonlydefaults[arg]
                    if arg_spec.kwonlydefaults is not None and arg in arg_spec.kwonlydefaults
                    else PluginArgument.NO_DEFAULT_VALUE_SET
                ),
            )
            self.kwargs[arg] = argument
            self.all_args[arg] = argument

    def get_signature(self) -> str:
        """
        Generate the signature of this plugin.  The signature is a string representing the the function
        as it can be called as a plugin in the model.
        """
        # Start the list with all positional arguments
        arg_list = [str(arg) for arg in self.args]

        # Filter all positional arguments out of the kwargs list
        kwargs = [str(arg) for _, arg in self.kwargs.items() if arg.is_kw_only_argument]

        if self.var_args is not None:
            arg_list.append("*" + str(self.var_args))
        elif kwargs:
            # For keyword only arguments, we need a marker if we don't have a catch-all
            # positional argument
            arg_list.append("*")

        # Add all keyword-only arguments to the list
        arg_list.extend(kwargs)

        if self.var_kwargs is not None:
            arg_list.append("**" + str(self.var_kwargs))

        # Join all arguments, separated by a comma
        args_string = ", ".join(arg_list)

        if self.return_type.type_expression is None:
            return "%s(%s)" % (self.__class__.__function_name__, args_string)
        return "%s(%s) -> %s" % (self.__class__.__function_name__, args_string, repr(self.return_type.type_expression))

    def get_arg(self, position: int) -> PluginArgument:
        """
        Get the argument at a given position, raise a RuntimeException if it doesn't exists..
        If a catch-all positional argument is defined, it will never raise a RuntimeException.

        :param position: The position of the argument (excluding any Context argument)
        """
        if position < len(self.args):
            return self.args[position]
        elif self.var_args is not None:
            return self.var_args
        else:
            raise RuntimeException(None, f"{self.get_full_name()}() got an unexpected positional argument: {position}")

    def get_kwarg(self, name: str) -> PluginArgument:
        """
        Get the argument with a given name, raise a RuntimeException if it doesn't exists.
        If a catch-all keyword argument is defined, it will never raise a RuntimeException.

        We currently don't support positional-only parameters, so we can simply look for any
        parameter named this way, positional or not.

        :param name: The name of the argument
        """
        if name in self.all_args:
            return self.all_args[name]
        elif self.var_kwargs is not None:
            return self.var_kwargs
        else:
            # Trying to provide a keyword argument which doesn't exist
            # The exception raised here tries to match as closely as possible what python
            # would have raised as exception
            raise RuntimeException(None, f"{self.get_full_name()}() got an unexpected keyword argument: '{name}'")

    def report_missing_arguments(
        self, missing_args: collections.abc.Sequence[str], args_sort: Literal["positional", "keyword-only"]
    ) -> None:
        """
        Helper method to raise an exception specifying that the given arguments are missing.  We try here
        to stick as much as possible to the error that python would have raised, only changing its type.

        The type of the exception raised is RuntimeException.

        If the list of missing_args is empty, we don't report any exception.

        :param missing_args: The missing arguments we should report
        :param args_sort: The sort of argument we are checking (positional or kw)
        """
        func = self.get_full_name()
        if len(missing_args) == 1:
            # The exception raised here tries to match as closely as possible what python
            # would have raised as exception
            raise RuntimeException(None, f"{func}() missing 1 required {args_sort} argument: '{missing_args[0]}'")
        if len(missing_args) > 1:
            arg_names = " and ".join(
                (
                    ", ".join(repr(arg) for arg in missing_args[:-1]),
                    repr(missing_args[-1]),
                )
            )
            # The exception raised here tries to match as closely as possible what python
            # would have raised as exception
            raise RuntimeException(None, f"{func}() missing {len(missing_args)} required {args_sort} arguments: {arg_names}")

    def check_args(self, args: Sequence[object], kwargs: Mapping[str, object]) -> bool:
        """
        Check if the arguments of the call match the function signature.

        1. Check if we have too many arguments
        2. Check if we have too few arguments
        3. Check if we have any duplicate arguments (provided as positional and keyword)
        4. Validate the type of each of the provided arguments

        :param args: All the positional arguments to pass on to the plugin function
        :param kwargs: All the keyword arguments to pass on to the plugin function
        """
        if self.var_args is None and len(args) > len(self.args):
            # (1) We got too many positional arguments
            # The exception raised here tries to match as closely as possible what python
            # would have raised as exception
            raise RuntimeException(
                None, f"{self.get_full_name()}() takes {len(self.args)} positional arguments but {len(args)} were given"
            )

        # (2) Check that all positional arguments without a default are provided
        missing_positional_arguments = [
            arg.arg_name
            for position, arg in enumerate(self.args)
            if (
                position >= len(args)  # No input from user in positional args
                and arg.arg_name not in kwargs  # No input from user in keyword args
                and not arg.has_default_value()  # No default value in plugin definition
            )
        ]
        self.report_missing_arguments(missing_positional_arguments, "positional")

        # (2) Check that all keyword arguments without a default are provided
        missing_keyword_arguments = [
            name
            for name, arg in self.kwargs.items()
            if (
                arg.is_kw_only_argument  # The argument was not yet checked as positional arg
                and name not in kwargs  # No input from user in keyword args
                and not arg.has_default_value()  # No default value in plugin definition
            )
        ]
        self.report_missing_arguments(missing_keyword_arguments, "keyword-only")

        # Validate all positional arguments
        for position, value in enumerate(args):
            # (1) Get the corresponding argument, fails if we don't have one
            arg = self.get_arg(position)

            # (4) Validate the input value
            if not arg.validate(value):
                return False

        # Validate all kw arguments
        for name, value in kwargs.items():
            # (1) Get the corresponding kwarg, fails if we don't have one
            kwarg = self.get_kwarg(name)

            # (3) Make sure that our argument is not provided twice
            if kwarg.arg_position is not None and kwarg.arg_position < len(args):
                # The exception raised here tries to match as closely as possible what python
                # would have raised as exception
                raise RuntimeException(None, f"{self.get_full_name()}() got multiple values for argument '{name}'")

            # (4) Validate the input value
            if not kwarg.validate(value):
                return False
        return True

    def emit_statement(self) -> "DynamicStatement":
        """
        This method is called to determine if the plugin call pushes a new
        statement
        """
        return self.new_statement

    def is_accept_unknowns(self) -> bool:
        return self.opts["allow_unknown"]

    def check_requirements(self) -> None:
        """
        Check if the plug-in has all it requires
        """
        if "bin" in self.opts and self.opts["bin"] is not None:
            for _bin in self.opts["bin"]:
                p = subprocess.Popen(["bash", "-c", "type -p %s" % _bin], stdout=subprocess.PIPE)
                result = p.communicate()

                if len(result[0]) == 0:
                    raise Exception(f"{self.__function_name__} requires {_bin} to be available in $PATH")

    @classmethod
    def deprecate_function(cls, replaced_by: Optional[str] = None) -> None:
        cls.deprecated = True
        cls.replaced_by = replaced_by

    def __call__(self, *args: object, **kwargs: object) -> object:
        """
        The function call itself
        """
        if self.deprecated:
            msg: str = f"Plugin '{self.get_full_name()}' is deprecated."
            if self.replaced_by:
                msg += f" It should be replaced by '{self.replaced_by}'."
            warnings.warn(PluginDeprecationWarning(msg))
        self.check_requirements()

        def new_arg(arg: object) -> object:
            if isinstance(arg, Context):
                return arg
            elif isinstance(arg, Unknown) and self.is_accept_unknowns():
                return arg
            else:
                return DynamicProxy.return_value(arg)

        new_args = [new_arg(arg) for arg in args]
        new_kwargs = {k: new_arg(v) for k, v in kwargs.items()}

        value = self.call(*new_args, **new_kwargs)

        value = DynamicProxy.unwrap(value)

        # Validate the returned value
        self.return_type.validate(value)

        return value

    def get_full_name(self) -> str:
        return f"{self.ns.get_full_name()}::{self.__class__.__function_name__}"

    def type_string(self) -> str:
        return self.get_full_name()


@stable_api
class PluginException(Exception):
    """
    Base class for custom exceptions raised from a plugin.
    """

    def __init__(self, message: str) -> None:
        self.message = message


@stable_api
def plugin(
    function: Optional[Callable] = None,
    commands: Optional[list[str]] = None,
    emits_statements: bool = False,
    allow_unknown: bool = False,
) -> Callable:
    """
    Python decorator to register functions with inmanta as plugin

    :param function: The function to register with inmanta. This is the first argument when it is used as decorator.
    :param commands: A list of command paths that need to be available. Inmanta raises an exception when the command is
                     not available.
    :param emits_statements: Set to true if this plugin emits new statements that the compiler should execute. This is only
                             required for complex plugins such as integrating a template engine.
    :param allow_unknown: Set to true if this plugin accepts Unknown values as valid input.
    """

    def curry_name(
        name: Optional[str] = None,
        commands: Optional[list[str]] = None,
        emits_statements: bool = False,
        allow_unknown: bool = False,
    ) -> Callable:
        """
        Function to curry the name of the function
        """

        def call(fnc):
            """
            Create class to register the function and return the function itself
            """

            def wrapper(self, *args: object, **kwargs: object) -> Any:
                """
                Python will bind the function as method into the class
                """
                return fnc(*args, **kwargs)

            nonlocal name, commands, emits_statements

            if name is None:
                name = fnc.__name__

            ns_parts = str(fnc.__module__).split(".")
            ns_parts.append(name)
            if ns_parts[0] != const.PLUGINS_PACKAGE:
                raise Exception("All plugin modules should be loaded in the %s package" % const.PLUGINS_PACKAGE)

            fq_plugin_name = "::".join(ns_parts[1:])

            dictionary = {}
            dictionary["__module__"] = fnc.__module__

            dictionary["__function_name__"] = name
            dictionary["__fq_plugin_name__"] = fq_plugin_name

            dictionary["opts"] = {"bin": commands, "emits_statements": emits_statements, "allow_unknown": allow_unknown}
            dictionary["call"] = wrapper
            dictionary["__function__"] = fnc

            bases = (Plugin,)
            fnc.__plugin__ = PluginMeta.__new__(PluginMeta, name, bases, dictionary)
            return fnc

        return call

    if function is None:
        return curry_name(commands=commands, emits_statements=emits_statements, allow_unknown=allow_unknown)

    elif isinstance(function, str):
        return curry_name(function, commands=commands, emits_statements=emits_statements, allow_unknown=allow_unknown)

    elif function is not None:
        fnc = curry_name(commands=commands, emits_statements=emits_statements, allow_unknown=allow_unknown)
        return fnc(function)


@stable_api
def deprecated(
    function: Optional[Callable] = None, *, replaced_by: Optional[str] = None, **kwargs: abc.Mapping[str, object]
) -> Callable:
    """
    the kwargs are currently ignored but where added in case we want to add something later on.
    """

    def inner(fnc: Callable):
        if hasattr(fnc, "__plugin__"):
            fnc.__plugin__.deprecate_function(replaced_by)
        else:
            raise Exception(
                f"Can not deprecate '{fnc.__name__}': The '@deprecated' decorator should be used in combination with the "
                f"'@plugin' decorator and should be placed at the top."
            )
        return fnc

    if function is not None:
        return inner(function)
    return inner



"""
    Copyright 2017 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

import logging
import os
import platform
from typing import TYPE_CHECKING, Callable, Union

if TYPE_CHECKING:
    from inmanta.agent.agent import Agent

try:
    import resource
except ImportError:
    resource = None

LOGGER = logging.getLogger(__name__)

ReportReturn = Union[dict[str, list[str]], dict[str, str], dict[str, float], str]
reports: dict[str, Callable[["Agent"], ReportReturn]] = {}


def collect_report(agent: "Agent") -> dict[str, ReportReturn]:
    out = {}
    for name, report in reports.items():
        try:
            out[name] = report(agent)
        except Exception:
            out[name] = "ERROR"
            LOGGER.exception("could generate report for entry: %s" % name)

    return out


def report_environment(agent: "Agent") -> str:
    return str(agent._env_id)


reports["environment"] = report_environment


def report_platform(agent: "Agent") -> str:
    value = platform.platform()
    if value is None:
        return "unknown"
    return value


reports["platform"] = report_platform


def report_hostname(agent: "Agent") -> str:
    return platform.node()


reports["hostname"] = report_hostname


def report_python(agent: "Agent") -> str:
    return f"{platform.python_implementation()} {platform.python_version()} {platform.python_build()}"


reports["python"] = report_python


def report_pid(agent: "Agent") -> str:
    return str(os.getpid())


reports["pid"] = report_pid


def report_resources(agent: "Agent") -> dict[str, float]:
    if resource is None:
        return {}

    ru = resource.getrusage(resource.RUSAGE_SELF)
    out = {
        "utime": ru.ru_utime,
        "stime": ru.ru_stime,
        "maxrss": ru.ru_maxrss,
        "ixrss": ru.ru_ixrss,
        "idrss": ru.ru_idrss,
        "isrss": ru.ru_isrss,
        "minflt": ru.ru_minflt,
        "majflt": ru.ru_majflt,
        "nswap": ru.ru_nswap,
        "inblock": ru.ru_inblock,
        "oublock": ru.ru_oublock,
        "msgsnd": ru.ru_msgsnd,
        "msgrcv": ru.ru_msgrcv,
        "nsignals": ru.ru_nsignals,
        "nvcsw": ru.ru_nvcsw,
        "nivcsw": ru.ru_nivcsw,
    }
    return out


reports["resources"] = report_resources



"""
    Copyright 2017 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

# pylint: disable-msg=W0613
import typing
from collections import abc
from itertools import chain
from string import Formatter
from typing import Optional, TypeVar

import inmanta.execute.dataflow as dataflow
from inmanta.ast import (
    AttributeException,
    DuplicateException,
    HyphenException,
    KeyException,
    LocatableString,
    Location,
    OptionalValueException,
    RuntimeException,
    TypeAnchor,
    TypingException,
)
from inmanta.ast.attribute import RelationAttribute
from inmanta.ast.statements import (
    AssignStatement,
    AttributeAssignmentLHS,
    ExpressionStatement,
    RequiresEmitStatement,
    Resumer,
    Statement,
    StaticEagerPromise,
)
from inmanta.execute.dataflow import DataflowGraph
from inmanta.execute.runtime import (
    ExecutionUnit,
    HangUnit,
    Instance,
    ListLiteral,
    QueueScheduler,
    Resolver,
    ResultCollector,
    ResultVariable,
    VariableABC,
)
from inmanta.execute.util import Unknown

from . import ReferenceStatement

try:
    from typing import TYPE_CHECKING
except ImportError:
    TYPE_CHECKING = False


if TYPE_CHECKING:
    from inmanta.ast.statements.generator import WrappedKwargs  # noqa: F401
    from inmanta.ast.variables import Reference  # noqa: F401

T = TypeVar("T")


class CreateList(ReferenceStatement):
    """
    Represents a list literal statement which might contain any type of value (constants and/or instances).
    """

    __slots__ = ("items",)

    def __init__(self, items: list[ExpressionStatement]) -> None:
        ReferenceStatement.__init__(self, items)
        self.items = items

    def normalize(self, *, lhs_attribute: Optional[AttributeAssignmentLHS] = None) -> None:
        for item in self.items:
            # pass on lhs_attribute to children
            item.normalize(lhs_attribute=lhs_attribute)
            self.anchors.extend(item.get_anchors())

    def requires_emit_gradual(
        self, resolver: Resolver, queue: QueueScheduler, resultcollector: Optional[ResultCollector]
    ) -> dict[object, VariableABC]:
        if resultcollector is None:
            return self.requires_emit(resolver, queue)

        requires: dict[object, VariableABC] = self._requires_emit_promises(resolver, queue)

        # if we are in gradual mode, transform to a list of assignments instead of assignment of a list
        # to get more accurate gradual execution
        # temp variable is required get all heuristics right

        # ListVariable to hold all the stuff. Used as a proxy for gradual execution and to track promises.
        # Freezes itself once all promises have been fulfilled, at which point it represents the full list literal created by
        # this statement.
        temp = ListLiteral(queue)

        # add listener for gradual execution
        temp.listener(resultcollector, self.location)

        # Assignments, wired for gradual
        for expr in self.items:
            ExecutionUnit(queue, resolver, temp, expr.requires_emit_gradual(resolver, queue, temp), expr, self)

        if not self.items:
            # empty: just close
            temp.freeze()

        # pass temp
        requires[self] = temp
        return requires

    def _resolve(self, requires: dict[object, object], resolver: Resolver, queue: QueueScheduler) -> object:
        """
        Create this list
        """
        # gradual case, everything is in placeholder
        if self in requires:
            return requires[self]

        qlist = []

        for i in range(len(self.items)):
            value = self.items[i].execute(requires, resolver, queue)
            if isinstance(value, list):
                qlist.extend(value)
            else:
                qlist.append(value)

        return qlist

    def execute_direct(self, requires: abc.Mapping[str, object]) -> object:
        qlist = []
        for i in range(len(self.items)):
            value = self.items[i]
            item = value.execute_direct(requires)
            if isinstance(item, list):
                # flatten cfr BaseListVariable._set_value
                qlist.extend(item)
            else:
                qlist.append(item)

        return qlist

    def as_constant(self) -> list[object]:
        list_result = (v.as_constant() for v in self.items)
        # flatten cfr BaseListVariable._set_value
        flat_result = [item for element in list_result for item in ((element,) if not isinstance(element, list) else element)]
        return flat_result

    def get_dataflow_node(self, graph: DataflowGraph) -> dataflow.NodeReference:
        return dataflow.NodeStub("CreateList.get_node() placeholder for %s" % self).reference()

    def pretty_print(self) -> str:
        return "[%s]" % ",".join(item.pretty_print() for item in self.items)

    def __repr__(self) -> str:
        return "List()"


class CreateDict(ReferenceStatement):
    __slots__ = ("items",)

    def __init__(self, items: list[tuple[str, ReferenceStatement]]) -> None:
        ReferenceStatement.__init__(self, [x[1] for x in items])
        self.items = items
        seen = {}  # type: typing.Dict[str,ReferenceStatement]
        for x, v in items:
            if x in seen:
                raise DuplicateException(v, seen[x], "duplicate key in dict %s" % x)
            seen[x] = v

    def execute_direct(self, requires: abc.Mapping[str, object]) -> object:
        qlist = {}

        for i in range(len(self.items)):
            key, value = self.items[i]
            qlist[key] = value.execute_direct(requires)

        return qlist

    def _resolve(self, requires: dict[object, object], resolver: Resolver, queue: QueueScheduler) -> object:
        """
        Create this list
        """
        qlist = {}

        for i in range(len(self.items)):
            key, value = self.items[i]
            qlist[key] = value.execute(requires, resolver, queue)

        return qlist

    def as_constant(self) -> dict[str, object]:
        return {k: v.as_constant() for k, v in self.items}

    def get_dataflow_node(self, graph: DataflowGraph) -> dataflow.NodeReference:
        return dataflow.NodeStub("CreateDict.get_node() placeholder for %s" % self).reference()

    def __repr__(self) -> str:
        return "Dict()"


class SetAttribute(AssignStatement, Resumer):
    """
    Set an attribute of a given instance to a given value
    """

    __slots__ = ("instance", "attribute_name", "value", "list_only", "_assignment_promise")

    def __init__(self, instance: "Reference", attribute_name: str, value: ExpressionStatement, list_only: bool = False) -> None:
        AssignStatement.__init__(self, instance, value)
        self.instance = instance
        self.attribute_name = attribute_name
        self.value = value
        self.list_only = list_only
        self._assignment_promise: StaticEagerPromise = StaticEagerPromise(self.instance, self.attribute_name, self)

    def normalize(self, *, lhs_attribute: Optional[AttributeAssignmentLHS] = None) -> None:
        # register this assignment as left hand side to the value on the right hand side
        self.rhs.normalize(lhs_attribute=AttributeAssignmentLHS(self.instance, self.attribute_name))
        self.anchors.extend(self.rhs.get_anchors())
        if self.lhs is not None:
            self.anchors.extend(self.lhs.get_anchors())

    def get_all_eager_promises(self) -> abc.Iterator["StaticEagerPromise"]:
        # propagate this attribute assignment's promise to parent blocks
        return chain(super().get_all_eager_promises(), [self._assignment_promise])

    def _add_to_dataflow_graph(self, graph: typing.Optional[DataflowGraph]) -> None:
        if graph is None:
            return
        node: dataflow.AttributeNodeReference = self.instance.get_dataflow_node(graph).get_attribute(self.attribute_name)
        node.assign(self.value.get_dataflow_node(graph), self, graph)

    def emit(self, resolver: Resolver, queue: QueueScheduler) -> None:
        self._add_to_dataflow_graph(resolver.dataflow_graph)
        reqs = self.instance.requires_emit(resolver, queue)
        # This class still implements custom attribute resolution, rather than using the new VariableReferenceHook mechanism
        HangUnit(queue, resolver, reqs, ResultVariable(), self)

    def resume(self, requires: dict[object, object], resolver: Resolver, queue: QueueScheduler, target: ResultVariable) -> None:
        instance = self.instance.execute(requires, resolver, queue)
        if not isinstance(instance, Instance):
            raise TypingException(
                self, f"The object at {self.instance} is not an Entity but a {type(instance)} with value {instance}"
            )
        var = instance.get_attribute(self.attribute_name)
        if self.list_only and not var.is_multi():
            raise TypingException(self, "Can not use += on relations with multiplicity 1")

        if var.is_multi():
            # gradual only for multi
            # to preserve order on lists used in attributes
            # while allowing gradual execution on relations
            reqs = self.value.requires_emit_gradual(
                resolver, queue, GradualSetAttributeHelper(self, instance, self.attribute_name, var)
            )
        else:
            reqs = self.value.requires_emit(resolver, queue)

        SetAttributeHelper(queue, resolver, var, reqs, self.value, self, instance, self.attribute_name)

    def pretty_print(self) -> str:
        return f"{self.instance.pretty_print()}.{self.attribute_name} = {self.value.pretty_print()}"

    def __str__(self) -> str:
        return f"{str(self.instance)}.{self.attribute_name} = {str(self.value)}"


class GradualSetAttributeHelper(ResultCollector[T]):
    """
    A result collector wrapper that ensures that exceptions that happen during assignment
    are attributed to the correct statement
    """

    __slots__ = ("stmt", "next", "instance", "attribute_name")

    def __init__(self, stmt: "Statement", instance: "Instance", attribute_name: str, next: ResultCollector[T]) -> None:
        self.stmt = stmt
        self.instance = instance
        self.next = next
        self.attribute_name = attribute_name

    def receive_result(self, value: T, location: Location) -> bool:
        try:
            self.next.receive_result(value, location)
            return False
        except AttributeException as e:
            e.set_statement(self.stmt, False)
            raise
        except RuntimeException as e:
            e.set_statement(self.stmt, False)
            raise AttributeException(self.stmt, self.instance, self.attribute_name, e)


class SetAttributeHelper(ExecutionUnit):
    __slots__ = ("stmt", "instance", "attribute_name")

    def __init__(
        self,
        queue_scheduler: QueueScheduler,
        resolver: Resolver,
        result: ResultVariable,
        requires: dict[object, ResultVariable],
        expression: ExpressionStatement,
        stmt: Statement,
        instance: Instance,
        attribute_name: str,
    ) -> None:
        ExecutionUnit.__init__(self, queue_scheduler, resolver, result, requires, expression)
        self.stmt = stmt
        self.instance = instance
        self.attribute_name = attribute_name

    def execute(self) -> None:
        try:
            ExecutionUnit._unsafe_execute(self)
        except AttributeException as e:
            e.set_statement(self.stmt, False)
            raise
        except OptionalValueException as e:
            # OptionalValueException has only its instance as statement, override with more accurate statement and location
            e.set_statement(self.stmt, True)
            e.location = self.stmt.location
            raise AttributeException(self.stmt, self.instance, self.attribute_name, e)
        except RuntimeException as e:
            e.set_statement(self.stmt, False)
            raise AttributeException(self.stmt, self.instance, self.attribute_name, e)

    def __str__(self) -> str:
        return str(self.stmt)


class Assign(AssignStatement):
    """
    This class represents the assignment of a value to a variable -> alias

    :param name: The name of the value
    :param value: The value that is to be assigned to the variable

    uses:          value
    provides:      variable
    """

    def __init__(self, name: LocatableString, value: ExpressionStatement) -> None:
        AssignStatement.__init__(self, None, value)
        self.name = name
        self.value = value
        if "-" in str(self.name):
            raise HyphenException(name)

    def _add_to_dataflow_graph(self, graph: typing.Optional[DataflowGraph]) -> None:
        if graph is None:
            return
        node: dataflow.AssignableNodeReference = graph.resolver.get_dataflow_node(str(self.name))
        node.assign(self.value.get_dataflow_node(graph), self, graph)

    def emit(self, resolver: Resolver, queue: QueueScheduler) -> None:
        self._add_to_dataflow_graph(resolver.dataflow_graph)
        target = resolver.lookup(str(self.name))
        assert isinstance(target, ResultVariable)
        reqs = self.value.requires_emit(resolver, queue)
        ExecutionUnit(queue, resolver, target, reqs, self.value, owner=self)

    def declared_variables(self) -> abc.Iterator[str]:
        yield str(self.name)

    def pretty_print(self) -> str:
        return f"{self.name} = {self.value.pretty_print()}"

    def __repr__(self) -> str:
        return f"Assign({self.name}, {self.value})"

    def __str__(self) -> str:
        return f"{self.name} = {self.value}"


class MapLookup(ReferenceStatement):
    """
    Lookup a value in a dict
    """

    __slots__ = ("themap", "key", "location")

    def __init__(self, themap: ExpressionStatement, key: ExpressionStatement):
        super().__init__([themap, key])
        self.themap = themap
        self.key = key
        self.location = themap.get_location().merge(key.location)

    def _resolve(self, requires: dict[object, object], resolver: Resolver, queue: QueueScheduler) -> object:
        mapv = self.themap.execute(requires, resolver, queue)
        if isinstance(mapv, Unknown):
            return Unknown(self)
        if not isinstance(mapv, dict):
            raise TypingException(self, "dict lookup is only possible on dicts, %s is not an object" % mapv)

        keyv = self.key.execute(requires, resolver, queue)
        if isinstance(keyv, Unknown):
            return Unknown(self)
        if not isinstance(keyv, str):
            raise TypingException(self, "dict keys must be string, %s is not a string" % keyv)

        if keyv not in mapv:
            raise KeyException(self, "key {} not found in dict, options are [{}]".format(keyv, ",".join(mapv.keys())))

        return mapv[keyv]

    def get_dataflow_node(self, graph: DataflowGraph) -> dataflow.NodeReference:
        return dataflow.NodeStub("MapLookup.get_node() placeholder for %s" % self).reference()

    def __repr__(self) -> str:
        return f"{repr(self.themap)}[{repr(self.key)}]"


class IndexLookup(ReferenceStatement, Resumer):
    """
    Lookup a value in a dictionary
    """

    __slots__ = ("index_type", "query", "wrapped_query", "type")

    def __init__(
        self,
        index_type: LocatableString,
        query: list[tuple[LocatableString, ExpressionStatement]],
        wrapped_query: list["WrappedKwargs"],
    ) -> None:
        ReferenceStatement.__init__(self, list(chain((v for (_, v) in query), wrapped_query)))
        self.index_type = index_type
        self.query = [(str(n), e) for n, e in query]
        self.wrapped_query: list["WrappedKwargs"] = wrapped_query

    def normalize(self, *, lhs_attribute: Optional[AttributeAssignmentLHS] = None) -> None:
        ReferenceStatement.normalize(self)
        self.type = self.namespace.get_type(self.index_type)
        self.anchors.append(TypeAnchor(self.index_type, self.type))

    def requires_emit(self, resolver: Resolver, queue: QueueScheduler) -> dict[object, VariableABC]:
        requires: dict[object, VariableABC] = RequiresEmitStatement.requires_emit(self, resolver, queue)
        sub = ReferenceStatement.requires_emit(self, resolver, queue)
        temp = ResultVariable()
        temp.set_type(self.type)
        HangUnit(queue, resolver, sub, temp, self)
        requires[self] = temp
        return requires

    def resume(self, requires: dict[object, object], resolver: Resolver, queue: QueueScheduler, target: ResultVariable) -> None:
        self.type.lookup_index(
            list(
                chain(
                    ((k, v.execute(requires, resolver, queue)) for (k, v) in self.query),
                    ((k, v) for kwargs in self.wrapped_query for (k, v) in kwargs.execute(requires, resolver, queue)),
                )
            ),
            self,
            target,
        )

    def _resolve(self, requires: dict[object, object], resolver: Resolver, queue: QueueScheduler) -> object:
        return requires[self]

    def get_dataflow_node(self, graph: DataflowGraph) -> dataflow.NodeReference:
        return dataflow.NodeStub("IndexLookup.get_node() placeholder for %s" % self).reference()

    def __repr__(self) -> str:
        """
        The representation of this statement
        """
        return "{}[{}]".format(self.index_type, ",".join([repr(x) for x in chain([self.query], self.wrapped_query)]))


class ShortIndexLookup(IndexLookup):
    """lookup of the form
    entity ExtendedHost extends ip::Host:
    end

    ExtendedHost.resources [0:] -- std::testing::NullResource.host[1]

    host = ExtendedHost(...)
    resource = std::testing::NullResource(name='test', host=host, ...)

    index std::testing::NullResource(host, name)

    host.resources[name="test"]
    """

    __slots__ = ("rootobject", "relation", "querypart", "wrapped_querypart")

    def __init__(
        self,
        rootobject: ExpressionStatement,
        relation: LocatableString,
        query: list[tuple[LocatableString, ExpressionStatement]],
        wrapped_query: list["WrappedKwargs"],
    ):
        ReferenceStatement.__init__(self, list(chain((v for (_, v) in query), [rootobject], wrapped_query)))
        self.rootobject = rootobject
        self.relation = str(relation)
        self.querypart: list[tuple[str, ExpressionStatement]] = [(str(n), e) for n, e in query]
        self.wrapped_querypart: list["WrappedKwargs"] = wrapped_query

    def normalize(self, *, lhs_attribute: Optional[AttributeAssignmentLHS] = None) -> None:
        ReferenceStatement.normalize(self)
        # currently there is no way to get the type of an expression prior to evaluation
        self.type = None

    def resume(self, requires: dict[object, object], resolver: Resolver, queue: QueueScheduler, target: ResultVariable) -> None:
        root_object = self.rootobject.execute(requires, resolver, queue)

        if not isinstance(root_object, Instance):
            raise TypingException(self, "short index lookup is only possible one objects, %s is not an object" % root_object)

        from_entity = root_object.get_type()

        relation = from_entity.get_attribute(self.relation)
        if not isinstance(relation, RelationAttribute):
            raise TypingException(self, "short index lookup is only possible on relations, %s is an attribute" % relation)

        if relation.end is None:
            raise TypingException(
                self, "short index lookup is only possible on bi-drectional relations, %s is unidirectional" % relation
            )

        self.type = relation.type

        self.type.lookup_index(
            list(
                chain(
                    [(relation.end.name, root_object)],
                    ((k, v.execute(requires, resolver, queue)) for (k, v) in self.querypart),
                    ((k, v) for kwargs in self.wrapped_querypart for (k, v) in kwargs.execute(requires, resolver, queue)),
                )
            ),
            self,
            target,
        )

    def __repr__(self) -> str:
        """
        The representation of this statement
        """
        return "{}.{}[{}]".format(
            self.rootobject,
            self.relation,
            ",".join(repr(part) for part in chain([self.querypart], self.wrapped_querypart)),
        )


class FormattedString(ReferenceStatement):
    """
    This class is an abstraction around a string containing references to variables.
    """

    __slots__ = ("_format_string",)

    def __init__(self, format_string: str, variables: abc.Sequence["Reference"]) -> None:
        super().__init__(variables)
        self._format_string = format_string

    def get_dataflow_node(self, graph: DataflowGraph) -> dataflow.NodeReference:
        return dataflow.NodeStub("StringFormat.get_node() placeholder for %s" % self).reference()

    def __repr__(self) -> str:
        return "Format(%r)" % self._format_string


class StringFormat(FormattedString):
    """
    Create a new string by doing a string interpolation
    """

    __slots__ = ("_variables",)

    def __init__(self, format_string: str, variables: abc.Sequence[tuple["Reference", str]]) -> None:
        super().__init__(format_string, [k for (k, _) in variables])
        self._variables = variables

    def _resolve(self, requires: dict[object, object], resolver: Resolver, queue: QueueScheduler) -> object:
        result_string = self._format_string
        for _var, str_id in self._variables:
            value = _var.execute(requires, resolver, queue)
            if isinstance(value, Unknown):
                return Unknown(self)
            if isinstance(value, float) and (value - int(value)) == 0:
                value = int(value)

            result_string = result_string.replace(str_id, str(value))

        return result_string


class FStringFormatter(Formatter):
    def __init__(self) -> None:
        Formatter.__init__(self)

    def get_field(self, key: str, args: abc.Sequence[object], kwargs: abc.Mapping[str, object]) -> tuple[object, str]:
        """
        Overrides Formatter.get_field. Composite variable names are expected to be resolved at this point and can be
        retrieved by their full name.

        Key is the full string between '{' and '}', e.g. ' a.b.c ' in '{ a.b.c }'. We override this method rather than get_value
        because we want to leverage the compiler to execute the reference, rather than just to execute the first component
        ('a'), and let Python handle subsequent components through attribute lookups, as is the default Formatter behavior.
        """
        # may raise KeyError, which has to be handled by format caller
        return (kwargs[key], key)


class StringFormatV2(FormattedString):
    """
    Create a new string by using python build in formatting

    """

    __slots__ = ("_variables",)

    def __init__(self, format_string: str, variables: abc.Sequence[tuple["Reference", str]]) -> None:
        """
        :param format_string: The string on which to perform substitution
        :param variables: Sequence of tuples each holding a normalized reference (i.e. stripped of eventual whitespaces ) to a
         variable to substitute in the format_string and the raw full name of this variable (i.e. including potential
         whitespaces).
        """
        only_refs: abc.Sequence["Reference"] = [k for (k, _) in variables]
        super().__init__(format_string, only_refs)
        self._variables: abc.Mapping[Reference, str] = {ref: full_name for (ref, full_name) in variables}

    def _resolve(self, requires: dict[object, object], resolver: Resolver, queue: QueueScheduler) -> object:
        formatter: FStringFormatter = FStringFormatter()

        # We can't cache the formatter because it has no ability to cache the parsed string

        kwargs = {}
        for _var, full_name in self._variables.items():
            value = _var.execute(requires, resolver, queue)
            if isinstance(value, Unknown):
                return Unknown(self)
            if isinstance(value, float) and (value - int(value)) == 0:
                value = int(value)
            kwargs[full_name] = value

        try:
            result_string = formatter.vformat(self._format_string, args=[], kwargs=kwargs)
        except KeyError as e:
            key: str = str(e)
            if key == "'0'":
                # special-case '{}' (which is valid in Python) with a more informative error message
                raise RuntimeException(
                    self, "f-strings do not support positional substitutions via '{}', use variable or attribute keys instead"
                )
            # this is probably not reachable in practice, but it might trigger if Python ever changes the '0' key
            # or the resolution order
            raise RuntimeException(self, f"Invalid key in f-string: '{key}'")
        except ValueError as e:
            raise RuntimeException(self, f"Invalid f-string: {e}")

        return result_string

    def __repr__(self) -> str:
        return "StringFormatV2(%r)" % self._format_string

    def __str__(self) -> str:
        return repr(self._format_string)



"""
    Copyright 2023 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

import abc
import collections.abc
import logging
import re
from dataclasses import dataclass
from types import TracebackType
from typing import Callable, NamedTuple, Optional, Type

from asyncpg import Connection

from inmanta.stable_api import stable_api

logger = logging.getLogger(__name__)

MODE_READ_COMMAND = 0
MODE_READ_INPUT = 1


class AsyncSingleton(collections.abc.AsyncIterable[bytes]):
    """AsyncPG wants an async iterable"""

    def __init__(self, item: bytes):
        self.item: Optional[bytes] = item

    def __aiter__(self) -> "AsyncSingleton":
        return self

    async def __anext__(self) -> bytes:
        if self.item is None:
            raise StopAsyncIteration
        item = self.item
        self.item = None
        return item


@stable_api
class PGRestore:
    """
    Class that offers support to restore a database dump.

    This class assumes that the names of schemas, tables and columns in the dump don't contain a dot, double quote or
    whitespace character.
    """

    PARSE_EXT_BUFFER_REGEX = re.compile(r"COPY (?P<fq_table_name>[^ ]+)[ ]+\((?P<columns>[^)]+)\)[ ]+FROM stdin")

    # asyncpg execute method can not read in COPY IN

    def __init__(self, script: list[str], postgresql_client: Connection) -> None:
        self.commandbuffer = ""
        self.extbuffer = ""
        self.mode = MODE_READ_COMMAND
        self.script = script
        self.client = postgresql_client

    async def run(self) -> None:
        for line in self.script:
            if self.mode == MODE_READ_COMMAND:
                if line.startswith("COPY"):
                    await self.execute_buffer()
                    self.extbuffer = line
                    self.mode = MODE_READ_INPUT
                else:
                    self.buffer(line)
            else:
                if line == "\\.\n":
                    await self.execute_input()
                    self.mode = MODE_READ_COMMAND
                else:
                    self.buffer(line)
        assert self.mode == MODE_READ_COMMAND
        await self.execute_buffer()

    def buffer(self, cmd: str) -> None:
        if cmd.startswith("--"):
            return
        if not cmd.strip():
            return
        self.commandbuffer += cmd

    async def execute_buffer(self) -> None:
        if not self.commandbuffer.strip():
            return
        await self.client.execute(self.commandbuffer)
        self.commandbuffer = ""

    async def _parse_fq_table_name(self, fq_table_name: str) -> tuple[Optional[str], str]:
        """
        Parse a fully qualified PostgreSQL table name into its schema and table components.

        :return: A tuple where the first element is the schema name and the second the table name.
                 If the provided fq_table_name doesn't contain a schema, the first element in the tuple
                 will be None.
        """
        if "." in fq_table_name:
            schema, table_name = fq_table_name.split(".", maxsplit=1)
        else:
            schema = None
            table_name = fq_table_name

        # The schema or table name might be surrounded in quotes when the name conflicts with a keyword.
        if schema:
            schema = schema.strip(' "')
        table_name = table_name.strip(' "')

        return schema, table_name

    async def _parse_copy_command_in_ext_buffer(self) -> tuple[Optional[str], str, list[str]]:
        assert self.extbuffer
        match = self.PARSE_EXT_BUFFER_REGEX.match(self.extbuffer)
        if match is None:
            raise Exception(f"Invalid COPY command: {self.extbuffer}")
        schema, table_name = await self._parse_fq_table_name(match.group("fq_table_name"))
        # A column name might be surrounded in quotes when the name conflicts with a keyword.
        columns = [elem.strip(' "') for elem in match.group("columns").split(",")]
        return schema, table_name, columns

    async def execute_input(self) -> None:
        schema_name, table_name, column_names = await self._parse_copy_command_in_ext_buffer()
        await self.client.copy_to_table(
            schema_name=schema_name,
            table_name=table_name,
            source=AsyncSingleton(self.commandbuffer.encode()),
            columns=column_names,
            timeout=10,
        )
        self.commandbuffer = ""


async def postgres_get_custom_types(postgresql_client: Connection) -> list[str]:
    """
    Returns all custom types defined in the database.
    """
    # Query extracted from CLI
    # psql -E
    # \dT

    get_custom_types = """
    SELECT n.nspname as "Schema",
      pg_catalog.format_type(t.oid, NULL) AS "Name",
      pg_catalog.obj_description(t.oid, 'pg_type') as "Description"
    FROM pg_catalog.pg_type t
         LEFT JOIN pg_catalog.pg_namespace n ON n.oid = t.typnamespace
    WHERE (t.typrelid = 0 OR (SELECT c.relkind = 'c' FROM pg_catalog.pg_class c WHERE c.oid = t.typrelid))
      AND NOT EXISTS(SELECT 1 FROM pg_catalog.pg_type el WHERE el.oid = t.typelem AND el.typarray = t.oid)
           AND n.nspname <> 'pg_catalog'
          AND n.nspname <> 'information_schema'
      AND pg_catalog.pg_type_is_visible(t.oid)
    ORDER BY 1, 2;
    """

    types_in_db = await postgresql_client.fetch(get_custom_types)
    type_names: list[str] = [str(x["Name"]) for x in types_in_db]

    return type_names


@stable_api
async def clear_database(postgresql_client: Connection) -> None:
    """
    Remove all content from the database. Removes functions, tables and data types.
    """
    assert not postgresql_client.is_in_transaction()
    await postgresql_client.reload_schema_state()
    # query taken from : https://database.guide/3-ways-to-list-all-functions-in-postgresql/
    functions_query = """
SELECT routine_name
FROM  information_schema.routines
WHERE routine_type = 'FUNCTION'
AND routine_schema = 'public';
    """
    functions_in_db = await postgresql_client.fetch(functions_query)
    function_names = [str(x["routine_name"]) for x in functions_in_db]
    if function_names:
        drop_query = "DROP FUNCTION if exists %s " % ", ".join(function_names)
        await postgresql_client.execute(drop_query)

    tables_in_db = await postgresql_client.fetch("SELECT table_name FROM information_schema.tables WHERE table_schema='public'")
    table_names = [f"public.{x['table_name']}" for x in tables_in_db]
    if table_names:
        drop_query = "DROP TABLE %s CASCADE" % ", ".join(table_names)
        await postgresql_client.execute(drop_query)

    type_names = await postgres_get_custom_types(postgresql_client)
    if type_names:
        drop_query = "DROP TYPE %s" % ", ".join(type_names)
        await postgresql_client.execute(drop_query)
    logger.info(
        "Performed Hard Clean with tables: %s  types: %s  functions: %s",
        ",".join(table_names),
        ",".join(type_names),
        ",".join(function_names),
    )


class ColumnDefinition(NamedTuple):
    """
    :param name: The name of the column.
    :param is_list: A boolean that indicates whether this column has the type list.
    :param default: The default value of this column. Or None, when this column doesn't have a default value.
    """

    name: str
    is_list: bool = False
    default: Optional[str] = None


@dataclass(frozen=True)
class EnumUpdateDefinition:
    """
    A definition on how an existing enum in the database has to be updated.

    :param name: The name of the enumeration.
    :param values: The values the enum should have after the update.
    :param deleted_values: A dictionary that indicates which elements are deleted from the existing enum and how
                           they should be migrated. The key of the dictionary is the name of the removed enum value
                           and the value of the dictionary is the value it should be replaced with or None if the new value
                           should be NULL.
    :param columns: A dictionary that indicates which columns of which tables are using the enum.
                    The key of the dictionary contains the name of the table.
    """

    name: str
    values: collections.abc.Sequence[str]
    deleted_values: collections.abc.Mapping[str, Optional[str]]
    columns: collections.abc.Mapping[str, collections.abc.Sequence[ColumnDefinition]]


async def replace_enum_type(new_type: EnumUpdateDefinition, *, connection: Connection) -> None:
    """
    Completely replaces an enum type with a new definition with the same name.

    :param new_type: The definition of the new type. Assumed to be an internal construct, this method is not safe against
                     injections via this object's attributes.
    """
    temp_name: str = f"_old_{new_type.name}"
    await connection.execute(
        f"""
        ALTER TYPE {new_type.name} RENAME TO {temp_name};
        CREATE TYPE {new_type.name} AS ENUM(%s);
        """
        % (", ".join(f"'{v}'" for v in new_type.values))
    )
    for table, columns in new_type.columns.items():
        for column, is_list, default in columns:
            for old_value, new_value in new_type.deleted_values.items():
                await connection.execute(f"UPDATE {table} SET {column}=$1 WHERE {column}=$2", new_value, old_value)
            await connection.execute(f"ALTER TABLE {table} ALTER COLUMN {column} DROP DEFAULT")
            if is_list:
                # can't cast directly between enums -> go via varchar
                await connection.execute(
                    f"ALTER TABLE {table} ALTER COLUMN {column} TYPE {new_type.name}[]"
                    f"USING {column}::varchar[]::{new_type.name}[]"
                )
            else:
                # can't cast directly between enums -> go via varchar
                await connection.execute(
                    f"ALTER TABLE {table} ALTER COLUMN {column} TYPE {new_type.name} USING {column}::varchar::{new_type.name}"
                )
            if default:
                await connection.execute(f"ALTER TABLE {table} ALTER COLUMN {column} SET DEFAULT '{default}'")
    await connection.execute(f"DROP TYPE {temp_name}")


class ConnectionMaybeInTransaction(abc.ABC):
    """A connection that is perhaps in a transaction"""

    def __init__(self, connection: Optional[Connection] = None) -> None:
        self.connection = connection

    @abc.abstractmethod
    def call_after_tx(self, finalizer: Callable[[], object]) -> None:
        """Add a method to be called after the transaction has committed successfully."""
        ...

    def __enter__(self) -> "ConnectionMaybeInTransaction":
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        return None


class ConnectionNotInTransaction(ConnectionMaybeInTransaction):
    """Connection that is not in a transaction or absent"""

    def call_after_tx(self, finalizer: Callable[[], object]) -> None:
        finalizer()


class ConnectionInTransaction(ConnectionMaybeInTransaction):
    def __init__(self, connection: Connection) -> None:
        super().__init__(connection)
        self.finished_callbacks: list[Callable[[], object]] = []

    def call_after_tx(self, finalizer: Callable[[], object]) -> None:
        self.finished_callbacks.append(finalizer)

    def __enter__(self) -> "ConnectionInTransaction":
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        if exc_type is None:
            for callback in self.finished_callbacks:
                callback()
        return None



"""
    Copyright 2019 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

import inspect
import json
import logging
import uuid
from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple, Type, cast, get_type_hints  # noqa: F401

import pydantic
import typing_inspect
from tornado import escape

from inmanta import const, util
from inmanta.data.model import BaseModel
from inmanta.protocol import auth, common, exceptions
from inmanta.protocol.common import ReturnValue
from inmanta.server import config as server_config
from inmanta.types import Apireturn, JsonType

LOGGER: logging.Logger = logging.getLogger(__name__)
INMANTA_MT_HEADER = "X-Inmanta-tid"

"""

RestServer  => manages tornado/handlers, marshalling, dispatching, and endpoints

ServerSlice => contributes handlers and methods

ServerSlice.server [1] -- RestServer.endpoints [1:]

"""


class CallArguments:
    """
    This class represents the call arguments for a method call.
    """

    def __init__(
        self, config: common.UrlMethod, message: dict[str, Optional[object]], request_headers: Mapping[str, str]
    ) -> None:
        """
        :param config: The method configuration that contains the metadata and functions to call
        :param message: The message received by the RPC call
        :param request_headers: The headers received by the RPC call
        """
        self._config = config
        self._properties = self._config.properties
        self._message = message
        self._request_headers = request_headers
        self._argspec: inspect.FullArgSpec = inspect.getfullargspec(self._properties.function)

        self._call_args: JsonType = {}
        self._headers: dict[str, str] = {}
        self._metadata: dict[str, object] = {}
        self._auth_token: Optional[auth.claim_type] = None
        self._auth_username: Optional[str] = None

        self._processed: bool = False

    @property
    def call_args(self) -> dict[str, object]:
        if not self._processed:
            raise Exception("Process call first before accessing property")

        return self._call_args

    @property
    def auth_username(self) -> Optional[str]:
        if not self._processed:
            raise Exception("Process call first before accessing property")

        return self._auth_username

    @property
    def headers(self) -> dict[str, str]:
        if not self._processed:
            raise Exception("Process call first before accessing property")

        return self._headers

    @property
    def metadata(self) -> dict[str, object]:
        if not self._processed:
            raise Exception("Process call first before accessing property")

        return self._metadata

    def _is_header_param(self, arg: str) -> bool:
        """
        Return True if the given call argument can be provided using a header parameter.
        """
        if arg not in self._properties.arg_options:
            return False

        opts = self._properties.arg_options[arg]
        return opts.header is not None

    def _is_header_param_provided(self, arg: str) -> bool:
        """
        Return True iff a value for the given call argument was provided using a header parameter.

        :raise Exception: The given call argument is not a header parameter.
        """
        value = self._get_header_value_for(arg)
        return value is not None

    def _get_header_value_for(self, arg: str) -> Optional[str]:
        """
        Return the header value that belongs to the given call argument or None when the header was not set.

        :raise Exception: The given call argument is not a header parameter.
        """
        if not self._is_header_param(arg):
            raise Exception(f"Parameter {arg} is not a header parameter")
        opts = self._properties.arg_options[arg]
        return self._request_headers.get(opts.header)

    def _map_headers(self, arg: str) -> Optional[object]:
        if not self._is_header_param(arg):
            return None

        opts = self._properties.arg_options[arg]
        assert opts.header is not None

        value = self._request_headers.get(opts.header)
        if opts.reply_header and value is not None:
            self._headers[opts.header] = value

        return value

    def get_default_value(self, arg_name: str, arg_position: int, default_start: int) -> Optional[object]:
        """
        Get a default value for an argument
        """
        if self._argspec.defaults and default_start >= 0 and 0 <= (arg_position - default_start) < len(self._argspec.defaults):
            return self._argspec.defaults[arg_position - default_start]
        else:
            raise exceptions.BadRequest("Field '%s' is required." % arg_name)

    async def _run_getters(self, arg: str, value: Optional[object]) -> Optional[object]:
        """
        Run any available getters on value
        """
        if arg not in self._properties.arg_options or self._properties.arg_options[arg].getter is None:
            return value

        try:
            value = await self._properties.arg_options[arg].getter(value, self._metadata)
            return value
        except Exception as e:
            LOGGER.exception("Failed to use getter for arg %s", arg)
            raise e

    @staticmethod
    def _ensure_list_if_list_type(arg_type: Optional[Type[object]], arg_value: object) -> object:
        """
        Handles processing of arguments for GET requests, especially for list types encoded as URL query parameters.
        If a GET endpoint has a parameter of type list that is encoded as a URL query parameter and the specific request
        provides a list with one element, urllib doesn't parse it as a list. Map it here explicitly to a list.
        """
        if typing_inspect.is_optional_type(arg_type):
            non_none_arg_types = [arg for arg in typing_inspect.get_args(arg_type) if arg is not type(None)]
            if len(non_none_arg_types) == 1:
                arg_type = non_none_arg_types[0]

        is_generic_list = arg_type and typing_inspect.is_generic_type(arg_type) and typing_inspect.get_origin(arg_type) is list
        is_single_type_list = len(typing_inspect.get_args(arg_type, evaluate=True)) == 1
        arg_value_is_not_list = not isinstance(arg_value, list)

        if is_generic_list and is_single_type_list and arg_value_is_not_list:
            return [arg_value]
        return arg_value

    def _validate_argument_consistency(self, args):
        """
        Validates the consistency of arguments, ensuring they are not passed both as a header and a non-header value.
        """
        for arg in args:
            if arg in self._message and self._is_header_param(arg) and self._is_header_param_provided(arg):
                message_value = self._message[arg]
                header_value = self._get_header_value_for(arg)
                if message_value != header_value:
                    raise exceptions.BadRequest(
                        f"Value for argument {arg} was provided via a header and a non-header argument, but both values"
                        f" don't match (header={header_value}; non-header={message_value})"
                    )

    def get_call_context(self) -> Optional[str]:
        """Returns the name of the first handler argument that is of type CallContext"""
        for arg, hint in self._config.handler.__annotations__.items():
            if isinstance(hint, type) and issubclass(hint, common.CallContext):
                return arg
        return None

    async def process(self) -> None:
        """
        Process the request
        """
        args: list[str] = list(self._argspec.args)

        if "self" in args:
            args.remove("self")

        all_fields = set(self._message.keys())  # Track all processed fields to warn user
        defaults_start: int = -1
        if self._argspec.defaults is not None:
            defaults_start = len(args) - len(self._argspec.defaults)

        self._validate_argument_consistency(args)

        call_args = {}

        for i, arg in enumerate(args):
            arg_type: Optional[type[object]] = self._argspec.annotations.get(arg)
            if arg in self._message:
                # Argument is parameter in body of path of HTTP request
                value = self._message[arg]

                if arg_type and self._properties.operation == "GET":
                    value = self._ensure_list_if_list_type(arg_type, value)

                all_fields.remove(arg)

            elif arg_type and self._properties.operation == "GET" and self._is_dict_or_optional_dict(arg_type):
                # Argument is dictionary-based expression in query parameters of GET operation
                dict_prefix = f"{arg}."
                dict_with_prefixed_names = {
                    param_name: param_value
                    for param_name, param_value in self._message.items()
                    if param_name.startswith(dict_prefix) and len(param_name) > len(dict_prefix)
                }
                value = (
                    await self._get_dict_value_from_message(arg, dict_prefix, dict_with_prefixed_names)
                    if dict_with_prefixed_names
                    else None
                )

                for key in dict_with_prefixed_names.keys():
                    all_fields.remove(key)
            elif self._is_header_param(arg):
                value = self._map_headers(arg)
                if value is None:
                    value = self.get_default_value(arg, i, defaults_start)
            else:
                value = self.get_default_value(arg, i, defaults_start)

            call_args[arg] = value

        # validate types
        call_args = self._properties.validate_arguments(call_args)

        for arg, value in call_args.items():
            # run getters
            value = await self._run_getters(arg, value)

            self._call_args[arg] = value

        # discard session handling data
        if self._properties.agent_server and "sid" in all_fields:
            all_fields.remove("sid")

        if self._properties.varkw:
            # add all other arguments to the call args as well
            for field in all_fields:
                self._call_args[field] = self._message[field]

        if len(all_fields) > 0 and self._argspec.varkw is None:
            raise exceptions.BadRequest(
                "request contains fields %s that are not declared in method and no kwargs argument is provided." % all_fields
            )

        # rename arguments if the handler requests this
        if hasattr(self._config.handler, "__protocol_mapping__"):
            for k, v in self._config.handler.__protocol_mapping__.items():
                if v in self._call_args:
                    self._call_args[k] = self._call_args[v]
                    del self._call_args[v]

        # verify if we need to inject a CallContext
        if call_context_var := self.get_call_context():
            self._call_args[call_context_var] = common.CallContext(
                request_headers=self._headers, auth_token=self._auth_token, auth_username=self._auth_username
            )

        self._processed = True

    async def _get_dict_value_from_message(
        self, arg: str, dict_prefix: str, dict_with_prefixed_names: dict[str, object]
    ) -> dict[str, object]:
        value = {k[len(dict_prefix) :]: v for k, v in dict_with_prefixed_names.items()}
        # Check if the values should be converted to lists
        type_args = self._argspec.annotations.get(arg)
        if typing_inspect.is_optional_type(type_args):
            # If optional, get the type args from the not None type argument
            dict_args = typing_inspect.get_args(typing_inspect.get_args(type_args, evaluate=True)[0], evaluate=True)
        else:
            dict_args = typing_inspect.get_args(self._argspec.annotations.get(arg), evaluate=True)
        dict_value_arg_type = (
            typing_inspect.get_origin(dict_args[1]) if typing_inspect.get_origin(dict_args[1]) else dict_args[1]
        )
        if issubclass(dict_value_arg_type, list):
            value = {key: [val] if not isinstance(val, list) else val for key, val in value.items()}
        return value

    def _is_dict_or_optional_dict(self, arg_type: type[object]) -> bool:
        if typing_inspect.is_optional_type(arg_type):
            arg_type = typing_inspect.get_args(arg_type, evaluate=True)[0]
        arg_type = typing_inspect.get_origin(arg_type) if typing_inspect.get_origin(arg_type) else arg_type
        if typing_inspect.is_new_type(arg_type):
            arg_type = type(arg_type)
        return issubclass(arg_type, dict)

    def _validate_union_return(self, arg_type: type[object], value: object) -> None:
        """Validate a return with a union type
        :see: protocol.common.MethodProperties._validate_function_types
        """
        matching_type = None
        for t in typing_inspect.get_args(arg_type, evaluate=True):
            instanceof_type = t
            if typing_inspect.is_generic_type(t):
                instanceof_type = typing_inspect.get_origin(t)

            if isinstance(value, instanceof_type):
                if matching_type is not None:
                    raise exceptions.ServerError(
                        f"Return type is defined as a union {arg_type} for which multiple "
                        f"types match the provided value {value}"
                    )
                matching_type = t

        if matching_type is None:
            raise exceptions.BadRequest(
                f"Invalid return value, no matching type found in union {arg_type} for value type {type(value)}"
            )

        if typing_inspect.is_generic_type(matching_type):
            self._validate_generic_return(arg_type, matching_type)

    def _validate_generic_return(self, arg_type: type[object], value: object) -> None:
        """Validate List or Dict types.

        :note: we return any here because the calling function also returns any.
        """
        if issubclass(typing_inspect.get_origin(arg_type), list):
            if not isinstance(value, list):
                raise exceptions.ServerError(
                    f"Invalid return value, type needs to be a list. Argument type should be {arg_type}"
                )

            el_type = typing_inspect.get_args(arg_type, evaluate=True)[0]
            if el_type is Any:
                return
            for el in value:
                if typing_inspect.is_union_type(el_type):
                    self._validate_union_return(el_type, el)
                elif not isinstance(el, el_type):
                    raise exceptions.ServerError(f"Element {el} of returned list is not of type {el_type}.")

        elif issubclass(typing_inspect.get_origin(arg_type), dict):
            if not isinstance(value, dict):
                raise exceptions.ServerError(
                    f"Invalid return value, type needs to be a dict. Argument type should be {arg_type}"
                )

            el_type = typing_inspect.get_args(arg_type, evaluate=True)[1]
            if el_type is Any:
                return
            for k, v in value.items():
                if not isinstance(k, str):
                    raise exceptions.ServerError("Keys of return dict need to be strings.")

                if typing_inspect.is_union_type(el_type):
                    self._validate_union_return(el_type, v)
                if typing_inspect.is_generic_type(el_type):
                    self._validate_generic_return(el_type, v)
                elif not isinstance(v, el_type):
                    raise exceptions.ServerError(f"Element {v} of returned list is not of type {el_type}.")

        else:
            # This should not happen because of MethodProperties validation
            raise exceptions.BadRequest(
                f"Failed to validate generic type {arg_type} of return value, only List and Dict are supported"
            )

    async def process_return(self, result: Apireturn) -> common.Response:
        """A handler can return ApiReturn, so lets handle all possible return types and convert it to a Response

        Apireturn = Union[int, Tuple[int, Optional[JsonType]], "ReturnValue", "BaseModel"]
        """
        if "return" in self._argspec.annotations:  # new style with return type
            return_type = self._argspec.annotations["return"]

            if return_type is None:
                if result is not None:
                    raise exceptions.ServerError(
                        f"Method {self._config.method_name} returned a result but is defined as -> None"
                    )

                return common.Response.create(ReturnValue(status_code=200, response=None), envelope=False)

            # There is no obvious method to check if the return_type is a specific version of the generic ReturnValue
            # The way this is implemented in typing is different for python 3.6 and 3.7. In this code we "trust" that the
            # signature of the handler and the method definition matches and the returned value matches this return value
            # Both isubclass and isinstance fail on this type
            # This check needs to be first because isinstance fails on generic types.
            # TODO: also validate the value inside a ReturnValue
            if return_type is Any:
                return common.Response.create(
                    ReturnValue(response=result), self._config.properties.envelope, self._config.properties.envelope_key
                )

            elif typing_inspect.is_union_type(return_type):
                self._validate_union_return(return_type, result)
                return common.Response.create(
                    ReturnValue(response=result), self._config.properties.envelope, self._config.properties.envelope_key
                )

            if typing_inspect.is_generic_type(return_type):
                if isinstance(result, ReturnValue):
                    return common.Response.create(
                        result, self._config.properties.envelope, self._config.properties.envelope_key
                    )
                else:
                    self._validate_generic_return(return_type, result)
                    return common.Response.create(
                        ReturnValue(response=result), self._config.properties.envelope, self._config.properties.envelope_key
                    )

            elif isinstance(result, BaseModel):
                return common.Response.create(
                    ReturnValue(response=result), self._config.properties.envelope, self._config.properties.envelope_key
                )

            elif isinstance(result, common.VALID_SIMPLE_ARG_TYPES):
                return common.Response.create(
                    ReturnValue(response=result), self._config.properties.envelope, self._config.properties.envelope_key
                )

            else:
                raise exceptions.ServerError(
                    f"Method {self._config.method_name} returned an invalid result {result} "
                    "instead of a BaseModel or ReturnValue"
                )

        else:  # "old" style method definition
            if isinstance(result, tuple):
                if len(result) == 2:
                    code, body = result
                else:
                    raise exceptions.ServerError("Handlers for method call can only return a status code and a reply")

            elif isinstance(result, int):
                code = result
                body = None

            else:
                raise exceptions.ServerError(
                    f"Method {self._config.method_name} returned an invalid result {result} instead of a status code or tupple"
                )

            if body is not None:
                if self._config.properties.reply:
                    return common.Response.create(
                        ReturnValue(status_code=code, response=body),
                        self._config.properties.envelope,
                        self._config.properties.envelope_key,
                    )
                else:
                    LOGGER.warning("Method %s returned a result although it has no reply!")

            return common.Response.create(ReturnValue(status_code=code, response=None), envelope=False)

    def _parse_and_validate_auth_token(self) -> None:
        """Get the auth token provided by the caller and decode it.

        :return: A mapping of claims
        """
        header_value: Optional[str] = None

        if additional_header := server_config.server_additional_auth_header.get():
            if additional_header in self._request_headers:
                header_value = self._request_headers[additional_header]

        if header_value is None and "Authorization" in self._request_headers:
            # In Authorization it is parsed as a bearer token
            parts = self._request_headers["Authorization"].split(" ")

            if len(parts) != 2 or parts[0].lower() != "bearer":
                logging.getLogger(__name__).warning(
                    "Invalid JWT token header Authorization. A bearer token is expected, instead (%s was provided)",
                    header_value,
                )
                return None

            header_value = parts[1]

        if header_value is None:
            return None

        self._auth_token, cfg = auth.decode_token(header_value)

        if cfg.jwt_username_claim in self._auth_token:
            self._auth_username = str(self._auth_token[cfg.jwt_username_claim])

    def authenticate(self, auth_enabled: bool) -> None:
        """Fetch any identity information and authenticate. This will also load this authentication
        information in this instance.

        :param auth_enabled: is authentication enabled?
        """
        if not auth_enabled:
            return

        # get and validate the token. A valid token means that user is authenticated
        self._parse_and_validate_auth_token()
        if self._auth_token is None and self._config.properties.enforce_auth:
            # We only need a valid token when the endpoint enforces authentication
            raise exceptions.UnauthorizedException()

    def authorize_request(self, auth_enabled: bool) -> None:
        """Authorize a request based on the given data

        :param auth_enabled: is authentication enabled?
        """
        if not auth_enabled:
            return

        if self._auth_token is None:
            if self._config.properties.enforce_auth:
                # We only need a valid token when the endpoint enforces authentication and auth is enabled
                raise exceptions.UnauthorizedException()
            return None

        # Enforce environment restrictions
        env_key: str = const.INMANTA_URN + "env"
        if env_key in self._auth_token:
            if env_key not in self.metadata:
                raise exceptions.Forbidden("The authorization token is scoped to a specific environment.")

            if self.metadata[env_key] != "all" and self._auth_token[env_key] != self.metadata[env_key]:
                raise exceptions.Forbidden("The authorization token is not valid for the requested environment.")

        # Enforce client_types restrictions
        ok: bool = False
        ct_key: str = const.INMANTA_URN + "ct"
        for ct in self._auth_token[ct_key]:
            if ct in self._config.properties.client_types:
                ok = True

        if not ok:
            raise exceptions.Forbidden(
                "The authorization token does not have a valid client type for this call."
                + f" ({self._auth_token[ct_key]} provided, {self._config.properties.client_types} expected"
            )


# Shared
class RESTBase(util.TaskHandler[None]):
    """
    Base class for REST based client and servers
    """

    _id: str

    @property
    def id(self) -> str:
        return self._id

    def _decode(self, body: bytes) -> Optional[JsonType]:
        """
        Decode a response body
        """
        result = None
        if body is not None and len(body) > 0:
            result = cast(JsonType, json.loads(escape.to_basestring(body)))

        return result

    def validate_sid(self, sid: uuid.UUID) -> bool:
        raise NotImplementedError()

    async def _execute_call(
        self,
        config: common.UrlMethod,
        message: dict[str, object],
        request_headers: Mapping[str, str],
    ) -> common.Response:
        try:
            if config is None:
                raise Exception("This method is unknown! This should not occur!")

            if config.properties.validate_sid:
                if "sid" not in message or not isinstance(message["sid"], str):
                    raise exceptions.BadRequest("this is an agent to server call, it should contain an agent session id")

                sid = uuid.UUID(str(message["sid"]))
                if not isinstance(sid, uuid.UUID) or not self.validate_sid(sid):
                    raise exceptions.BadRequest("the sid %s is not valid." % message["sid"])

            # First check if the call is authenticated, then process the request so we can handle it and then authorize it.
            # Authorization might need data from the request but we do not want to process it before we are sure the call
            # is authenticated.
            arguments = CallArguments(config, message, request_headers)
            arguments.authenticate(server_config.server_enable_auth.get())
            await arguments.process()
            arguments.authorize_request(server_config.server_enable_auth.get())

            LOGGER.debug(
                "Calling method %s(%s) user=%s",
                config.method_name,
                ", ".join([f"{name}={common.shorten(str(value))}" for name, value in arguments.call_args.items()]),
                arguments.auth_username if arguments.auth_username else "<>",
            )

            result = await config.handler(**arguments.call_args)
            return await arguments.process_return(result)
        except pydantic.ValidationError:
            LOGGER.exception(f"The handler {config.handler} caused a validation error in a data model (pydantic).")
            raise exceptions.ServerError("data validation error.")

        except exceptions.BaseHttpException:
            LOGGER.debug("An HTTP Error occurred", exc_info=True)
            raise

        except Exception as e:
            LOGGER.exception("An exception occurred during the request.")
            raise exceptions.ServerError(str(e.args))



"""
    Copyright 2023 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

import enum
import logging
import logging.config
import os
import re
import sys
from argparse import Namespace
from collections import abc
from collections.abc import Iterator
from contextlib import contextmanager
from logging import handlers
from typing import Optional, TextIO

import colorlog
import yaml
from colorlog.formatter import LogColors

from inmanta import config, const
from inmanta.server import config as server_config
from inmanta.stable_api import stable_api

LOGGER = logging.getLogger(__name__)


def _is_on_tty() -> bool:
    return (hasattr(sys.stdout, "isatty") and sys.stdout.isatty()) or const.ENVIRON_FORCE_TTY in os.environ


def python_log_level_to_name(python_log_level: int) -> str:
    """Convert a python log level to a human readable version that works in log config files"""
    # we build the reverse mapping every time because
    # we don't want to use private fields
    # name_to_level = logging._levelToName
    # the underlying assumption is that this code is not performance critical
    name_to_level = logging.getLevelNamesMapping()
    level_to_name = {v: k for k, v in name_to_level.items()}

    result = level_to_name.get(python_log_level)
    if result is not None:
        return result
    return str(python_log_level)


"""
This dictionary maps the Inmanta log levels to the corresponding Python log levels
"""
log_levels = {
    "0": logging.ERROR,
    "1": logging.WARNING,
    "2": logging.INFO,
    "3": logging.DEBUG,
    "4": 2,
    "ERROR": logging.ERROR,
    "WARNING": logging.WARNING,
    "INFO": logging.INFO,
    "DEBUG": logging.DEBUG,
    "TRACE": 2,
}


@stable_api
class LoggingConfigExtension:
    """
    This class is used by an extension to declare the default logging config that should be used when
    no dict-based logging config was provided by the user. The FullLoggingConfig class provides support
    to merge this logging config into the main logging config of inmanta-core. This class supports only
    version 1 of Python's dictConfig format.
    """

    def __init__(
        self,
        *,
        formatters: Optional[abc.Mapping[str, object]] = None,
        handlers: Optional[abc.Mapping[str, object]] = None,
        loggers: Optional[abc.Mapping[str, object]] = None,
        root_handlers: Optional[abc.Set[str]] = None,
        log_dirs_to_create: Optional[abc.Set[str]] = None,
    ) -> None:
        """
        :param log_dirs_to_create: The log directories that should be created before the logging config can be used.
        """
        self.formatters = formatters if formatters else {}
        self.handlers = handlers if handlers else {}
        self.loggers = loggers if loggers else {}
        self.root_handlers = root_handlers if root_handlers else set()
        self.log_dirs_to_create = log_dirs_to_create if log_dirs_to_create else set()

    def ensure_log_dirs(self) -> None:
        """
        This method makes sure that the log directories, required by this logging config, are present on disk.
        The directories are created if they don't exist yet.
        """
        for directory in self.log_dirs_to_create:
            os.makedirs(directory, exist_ok=True)

    def validate_for_extension(self, extension_name: str) -> None:
        """
        Verify that the names of the formatters and handlers are prefixed with `<name-extension>_` and
        raise an Exception in case this constraint is violated.
        """
        for logging_config_element in ["formatters", "handlers"]:
            for name in getattr(self, logging_config_element):
                if not name.startswith(f"{extension_name}_"):
                    raise Exception(
                        f"{logging_config_element.capitalize()} defined in the default logging config of an extension must be"
                        f" prefixed with `{extension_name}_`. Extension {extension_name} defines a"
                        f" {logging_config_element[0:-1]} with the invalid name {name}."
                    )


class FullLoggingConfig(LoggingConfigExtension):
    """
    A FullLoggingConfig that can be applied on Python's logging framework.

    This class supports only version 1 of Python's dictConfig format.
    """

    def __init__(
        self,
        *,
        formatters: Optional[abc.Mapping[str, object]] = None,
        handlers: Optional[abc.Mapping[str, object]] = None,
        loggers: Optional[abc.Mapping[str, object]] = None,
        root_handlers: Optional[abc.Set[str]] = None,
        log_dirs_to_create: Optional[abc.Set[str]] = None,
        root_log_level: Optional[int | str] = None,
    ):
        super().__init__(
            formatters=formatters,
            handlers=handlers,
            loggers=loggers,
            root_handlers=root_handlers,
            log_dirs_to_create=log_dirs_to_create,
        )
        self.root_log_level = root_log_level

    def join(self, logging_config_extension: LoggingConfigExtension) -> "FullLoggingConfig":
        """
        Join this FullLoggingConfig and the given LoggingConfigExtension together into a single
        FullLoggingConfig object that contains all the logging config of both objects.
        """
        common_formatter_names = set(self.formatters.keys()) & set(logging_config_extension.formatters.keys())
        if common_formatter_names:
            raise Exception(f"The following formatter names appear in multiple logging configs: {common_formatter_names}")
        common_handler_names = set(self.handlers.keys()) & set(logging_config_extension.handlers.keys())
        if common_handler_names:
            raise Exception(f"The following handler names appear in multiple logging configs: {common_handler_names}")
        common_logger_names = set(self.loggers.keys()) & set(logging_config_extension.loggers.keys())
        if common_logger_names:
            raise Exception(f"The following logger names appear in multiple logging configs: {common_logger_names}")
        return FullLoggingConfig(
            formatters=dict(**self.formatters, **logging_config_extension.formatters),
            handlers=dict(**self.handlers, **logging_config_extension.handlers),
            loggers=dict(**self.loggers, **logging_config_extension.loggers),
            root_handlers=set(*self.root_handlers, *logging_config_extension.root_handlers),
            log_dirs_to_create=set(*self.log_dirs_to_create, *logging_config_extension.log_dirs_to_create),
            root_log_level=self.root_log_level,
        )

    def apply_config(self) -> None:
        """
        Configure the logging system with this logging config.
        """
        self.ensure_log_dirs()
        dict_config = self._to_dict_config()
        logging.config.dictConfig(dict_config)

    def _to_dict_config(self) -> dict[str, object]:
        """
        Convert this object into a dictionary format that can be passed to logging.config.dictConfig() method
        to configure logging.
        """
        return {
            "version": 1,
            "formatters": dict(self.formatters),
            "handlers": dict(self.handlers),
            "loggers": dict(self.loggers),
            "root": {
                "handlers": self.root_handlers,
                **({"level": self.root_log_level} if self.root_log_level else {}),
            },
            "disable_existing_loggers": False,
        }


class Options(Namespace):
    """
    The Options class provides a way to configure the InmantaLoggerConfig with the following attributes:

    :param log_file: if this attribute is set, the logs will be written to the specified file instead of the stream
                     specified in `get_instance`.
    :param log_file_level: the Inmanta logging level for the file handler (if `log_file` is set).
                           The possible inmanta log levels and their associated python log level are defined in the
                           inmanta.logging.log_levels dictionary.
    :param verbose: the verbosity level of the log messages. can be a number from 0 to 4.
                    if a bigger number is provided, 4 will be used. Refer to log_file_level for the explanation of each level.
                    default is 1 (WARNING)
    :param timed: if true,  adds the time to the formatter in the log lines.
    :param logging_config: Path to the dict-based logging config file.
    """

    log_file: Optional[str] = None
    log_file_level: str = "INFO"
    verbose: int = 1
    timed: bool = False
    keep_logger_names: bool = False
    logging_config: Optional[str] = None


class LoggerMode(enum.Enum):
    """
    A different log format is used when the compiler/exporter is executed. This enum
    indicates which mode we are currently executing in.
        * COMPILER: the compiler is running.
        * EXPORT: The exporter is running.
        * OTHER: We are executing neither the compiler nor the exporter (e.g. running the server).
    """

    COMPILER = "compiler"
    EXPORTER = "exporter"
    OTHER = "other"


class LoggingConfigBuilder:

    def get_bootstrap_logging_config(
        self,
        stream: TextIO = sys.stdout,
        logging_config_extensions: Optional[abc.Sequence[LoggingConfigExtension]] = None,
        python_log_level: int = logging.INFO,
    ) -> FullLoggingConfig:
        """
        This method returns the logging config that should be used between the moment that the process starts,
        and the moment that the logging-related config options are parsed and applied.

        :param stream: The TextIO stream where the logs will be sent to.
        :param logging_config_extensions: The logging config required by the extensions.
        :param python_log_level: python log level to configure for the bootstrap logger
        """
        name_root_handler = "core_console_handler"
        log_level_name = python_log_level_to_name(python_log_level)
        logging_config_core = FullLoggingConfig(
            formatters={
                "core_console_formatter": self._get_multiline_formatter_config(),
            },
            handlers={
                name_root_handler: {
                    "class": "logging.StreamHandler",
                    "formatter": "core_console_formatter",
                    "level": log_level_name,
                    "stream": stream,
                },
            },
            loggers={},
            root_handlers={name_root_handler},
            root_log_level=log_level_name,
        )
        logging_config_core.validate_for_extension(extension_name="core")
        return self._join_logging_configs(logging_config_core, logging_config_extensions)

    def get_logging_config_from_options(
        self,
        stream: TextIO,
        options: Options,
        logging_config_extensions: Optional[abc.Sequence[LoggingConfigExtension]] = None,
    ) -> FullLoggingConfig:
        """
        Return the logging config based on the given configuration options, passed on the CLI,
        and the configuration options present in the config file.

        :param stream: The TextIO stream where the logs will be sent to.
        :param options: The config options passed on the CLI.
        :param logging_config_extensions: The logging config required by the extensions.
        """
        handlers: dict[str, object] = {}
        handler_root_logger: str
        log_level: int
        if options.log_file:
            log_level = convert_inmanta_log_level(options.log_file_level)
            handler_root_logger = "core_server_log"
            handlers[handler_root_logger] = {
                "class": "logging.handlers.WatchedFileHandler",
                "level": log_level,
                "formatter": "core_server_log_formatter",
                "filename": options.log_file,
                "mode": "a+",
            }
        else:
            log_level = convert_inmanta_log_level(inmanta_log_level=str(options.verbose), cli=True)
            handler_root_logger = "core_console"
            handlers[handler_root_logger] = {
                "class": "logging.StreamHandler",
                "formatter": "core_console_formatter",
                "level": log_level,
                "stream": stream,
            }

        full_logging_config = FullLoggingConfig(
            formatters={
                # Always add all the formatters, even if they are not used by configuration. This way
                # the formatters can be used if the user dumps the default logging config to file.
                "core_resource_action_log_formatter": {
                    "format": "%(asctime)s %(levelname)-8s %(name)-10s %(message)s",
                },
                "core_server_log_formatter": {
                    "format": "%(asctime)s %(levelname)-8s %(name)-10s %(message)s",
                },
                "core_console_formatter": self._get_multiline_formatter_config(options),
            },
            handlers={
                **handlers,
                "core_resource_action_handler": {
                    "class": "inmanta.logging.ParametrizedFileHandler",
                    "level": "DEBUG",
                    "formatter": "core_resource_action_log_formatter",
                    "name_parent_logger": const.NAME_RESOURCE_ACTION_LOGGER,
                    "log_file_template": os.path.join(
                        config.log_dir.get(), server_config.server_resource_action_log_prefix.get() + "{child_logger_name}.log"
                    ),
                },
                "core_tornado_debug_log_handler": {
                    "class": "inmanta.logging.TornadoDebugLogHandler",
                    "level": "DEBUG",
                },
            },
            loggers={
                const.NAME_RESOURCE_ACTION_LOGGER: {
                    "level": "DEBUG",
                    "propagate": True,
                    "handlers": ["core_resource_action_handler"],
                },
                "tornado.general": {
                    "level": "DEBUG",
                    "propagate": True,
                    "handlers": ["core_tornado_debug_log_handler"],
                },
            },
            root_handlers={handler_root_logger},
            root_log_level=log_level,
        )
        full_logging_config.validate_for_extension(extension_name="core")
        return self._join_logging_configs(full_logging_config, logging_config_extensions)

    def get_logging_config_for_agent(self, log_file: str, inmanta_log_level: str, cli_log: bool) -> FullLoggingConfig:
        """
        Returns the logging config for an agent.

        :param log_file: The log file were the logs should be sent to.
        :param inmanta_log_level: The Inmanta log level threshold, that indicates which log records should be ignored
                                  and which should be logged. This log level is taking into account for log records sent
                                  to file and to stderr.
        :param cli_log: A boolean indicating whether logs should also be sent to stderr or not.
        """
        python_log_level: int = convert_inmanta_log_level(inmanta_log_level)
        cli_handlers = {}
        name_root_handler = "core_agent_log_handler"
        root_loggers = {name_root_handler}
        if cli_log:
            cli_handlers["core_console_handler"] = {
                "class": "logging.StreamHandler",
                "formatter": "core_console_formatter",
                "level": python_log_level,
                "stream": "ext://sys.stderr",
            }
            root_loggers.add("core_console_handler")
        full_logging_config = FullLoggingConfig(
            formatters={
                # Always add all the formatters, even if they are not used by configuration. This way
                # the formatters can be used if the user dumps the default logging config to file.
                "core_agent_log_formatter": {
                    "format": "%(asctime)s %(levelname)-8s %(name)-10s %(message)s",
                },
                "core_console_formatter": self._get_multiline_formatter_config(),
            },
            handlers={
                name_root_handler: {
                    "class": "logging.handlers.WatchedFileHandler",
                    "level": python_log_level,
                    "formatter": "core_agent_log_formatter",
                    "filename": log_file,
                    "mode": "a+",
                },
                **cli_handlers,
            },
            loggers={},
            root_handlers=root_loggers,
            root_log_level=python_log_level,
        )
        full_logging_config.validate_for_extension(extension_name="core")
        return full_logging_config

    def _join_logging_configs(
        self,
        full_logger_config: FullLoggingConfig,
        logging_config_extensions: Optional[abc.Sequence[LoggingConfigExtension]] = None,
    ) -> FullLoggingConfig:
        """
        Join the given LoggingConfigCore and the LoggingConfigExtensions together into a single LoggingConfigCore
        object that contains all the loging config.
        """
        logging_config_extensions = logging_config_extensions if logging_config_extensions else []
        result = full_logger_config
        for logging_config_ext in logging_config_extensions:
            result = result.join(logging_config_ext)
        return result

    def _get_multiline_formatter_config(self, options: Optional[Options] = None) -> dict[str, object]:
        """
        Returns the dict-based formatter config for logs that will be sent to the console.

        :param options: The config options requested by the user or None if the config options are not parsed yet and
                        the bootstrap_logger_config should be used.
        """
        # Use a shorter space padding if we know that we will use short names as the logger name.
        # Otherwise the log records contains too much white spaces.
        space_padding_after_logger_name = (
            15
            if (
                options
                and not options.keep_logger_names
                and hasattr(options, "func")
                and options.func.__name__ in ["compile_project", "export"]
            )
            else 25
        )
        log_format = "%(asctime)s " if options and options.timed else ""
        if _is_on_tty():
            log_format += f"%(log_color)s%(name)-{space_padding_after_logger_name}s%(levelname)-8s%(reset)s%(blue)s%(message)s"
            log_colors = {"DEBUG": "cyan", "INFO": "green", "WARNING": "yellow", "ERROR": "red", "CRITICAL": "red"}
        else:
            log_format += f"%(name)-{space_padding_after_logger_name}s%(levelname)-8s%(message)s"
            log_colors = None

        return {
            "()": "inmanta.logging.MultiLineFormatter",
            "fmt": log_format,
            "log_colors": log_colors,
            "reset": _is_on_tty(),
            "no_color": not _is_on_tty(),
            "keep_logger_names": options.keep_logger_names if options else False,
        }


def convert_inmanta_log_level(inmanta_log_level: str, cli: bool = False) -> int:
    """
    Convert the given Inmanta log level to the corresponding Python log level.

    :param inmanta_log_level: The inmanta logging level
    :param cli: True if the logs will be outputted to the CLI.
    :return: python log level
    """
    # maximum of 4 v's
    if inmanta_log_level.isdigit() and int(inmanta_log_level) > 4:
        inmanta_log_level = "4"
    # The minimal log level on the CLI is always WARNING
    if cli and (inmanta_log_level == "ERROR" or (inmanta_log_level.isdigit() and int(inmanta_log_level) < 1)):
        inmanta_log_level = "WARNING"
    # Converts the Inmanta log level to the Python log level
    python_log_level = log_levels[inmanta_log_level]
    return python_log_level


class LoggerModeManager:
    """
    A singleton that keeps track of the current LoggerMode.
    """

    _instance: Optional["LoggerModeManager"] = None

    def __init__(self) -> None:
        self._logger_mode = LoggerMode.OTHER

    def get_logger_mode(self) -> LoggerMode:
        """
        Returns the current logger mode.
        """
        return self._logger_mode

    @contextmanager
    def run_in_logger_mode(self, logger_mode: LoggerMode) -> Iterator[None]:
        """
        A contextmanager that can be used to temporarily change the LoggerMode within a code block.
        This ContextManager updates the LoggerModeManager singleton and is therefore not async- or threadsafe.
        """
        prev_logger_mode = self._logger_mode
        self._logger_mode = logger_mode
        try:
            yield
        finally:
            self._logger_mode = prev_logger_mode

    @classmethod
    def get_instance(cls) -> "LoggerModeManager":
        if cls._instance is None:
            cls._instance = LoggerModeManager()
        return cls._instance


@stable_api
class InmantaLoggerConfig:
    """
    This class is the entry-point for configuring the Python logging framework.

    Usage:
    To use this class, you first need to call the `get_instance`. This method takes a `stream` argument
    that specifies where the log messages should be sent to. If no `stream` is provided,
    the log messages will be sent to standard output.

    You can then call the `apply_options` method to configure the logging options.

    The setup is not done in one step as we want logs for the cmd_parser, which will provide the options needed to configure
    the 'final' logger with `apply_options`.
    """

    _instance: Optional["InmantaLoggerConfig"] = None

    def __init__(self, stream: TextIO = sys.stdout) -> None:
        """
        Set up the logging handler for Inmanta

        :param stream: The TextIO stream where the logs will be sent to.
        """
        log_config: FullLoggingConfig = LoggingConfigBuilder().get_bootstrap_logging_config(stream)
        self._stream = stream
        self._handlers: abc.Sequence[logging.Handler] = self._apply_logging_config(log_config)
        self._options_applied: bool = False
        self._logging_configs_extensions: list[LoggingConfigExtension] = []

    @classmethod
    def get_current_instance(cls) -> "InmantaLoggerConfig":
        """
        Obtain the InmantaLoggerConfig singleton. This method assumes that an InmantaLoggerConfig was already initialized
        using the `get_instance()` method.
        """
        if not cls._instance:
            raise Exception("InmantaLoggerConfig was not yet initialized. Call get_instance() first.")
        return cls._instance

    @classmethod
    @stable_api
    def get_instance(cls, stream: TextIO = sys.stdout) -> "InmantaLoggerConfig":
        """
        This method should be used to obtain an instance of this class, because this class is a singleton.

        :param stream: The stream to send log messages to. Default is standard output (sys.stdout)
        """
        if cls._instance:
            if not cls._instance._handlers:
                raise Exception("No handlers found.")
            if not isinstance(cls._instance._handlers[0], logging.StreamHandler):
                raise Exception("Instance already exists with a different handler")
            elif isinstance(cls._instance._handlers[0], logging.StreamHandler) and cls._instance._handlers[0].stream != stream:
                raise Exception("Instance already exists with a different stream")
        else:
            cls._instance = cls(stream)
        return cls._instance

    @classmethod
    @stable_api
    def clean_instance(cls, root_handlers_to_remove: Optional[abc.Sequence[logging.Handler]] = None) -> None:
        """
        This method should be used to clean up an instance of this class.

        By default, this method removes and closes all root handlers from the logging framework. If the
        root_handlers_to_remove argument is not None, only the provided root handlers will be removed and closed.
        """
        to_remove = root_handlers_to_remove if root_handlers_to_remove is not None else logging.root.handlers
        for handler in to_remove:
            # File-based handlers automatically re-open after close() if log records are written to them.
            # As such, we explicitly remove the handler from the root logger here.
            logging.root.removeHandler(handler)
            handler.flush()
            handler.close()
        logging.shutdown()
        cls._instance = None

    def _get_path_to_logging_config_file(self, options: Options) -> Optional[str]:
        """
        Returns the path to the logging config file that was configured by the user
        or None if no logging config file was set.

        The configuration options are considered in the following order:
           1. The --logging-config CLI option.
           2. The INMANTA_CONFIG_LOGGING_CONFIG environment variable.
           3. The logging_config option in the config files.

        :param options: The configuration options passed on the CLI.
        """
        if options.logging_config:
            return options.logging_config
        return config.logging_config.get()

    @stable_api
    def apply_options(self, options: Options) -> None:
        """
        Apply the logging options to the current handler. A handler should have been created before

        :param options: The Option object coming from the command line. This function uses the following
            attributes: log_file, log_file_level, verbose, timed
        """
        if self._options_applied:
            raise Exception("Options can only be applied once to a handler.")

        logging_config_file: Optional[str] = self._get_path_to_logging_config_file(options)
        if logging_config_file:
            self._apply_logging_config_from_file(logging_config_file)
        else:
            self._apply_logging_config_from_options(options)
        self._options_applied = True

    def _apply_logging_config_from_file(self, config_file: str) -> None:
        """
        Apply the given logging config file.
        """
        handlers_before = list(logging.root.handlers)
        if not os.path.isfile(config_file):
            raise Exception(f"No logging config file exists at {config_file}.")
        with open(config_file, "r") as fh:
            try:
                dict_config = yaml.safe_load(fh)
            except yaml.YAMLError:
                raise Exception(f"Failed to parse the logging config file at {config_file} as YAML.")
        try:
            logging.config.dictConfig(dict_config)
        except Exception:
            raise Exception(f"Failed to apply the logging config defined in {dict_config}.")
        self._handlers = [handler for handler in logging.root.handlers if handler not in handlers_before]

    def _apply_logging_config_from_options(self, options: Options) -> None:
        """
        Apply the logging configuration as defined by the CLI options when the
        --logging-config option is not set.
        """
        config_builder = LoggingConfigBuilder()
        logging_config: FullLoggingConfig = config_builder.get_logging_config_from_options(
            self._stream, options, self._logging_configs_extensions
        )
        self._handlers = self._apply_logging_config(logging_config)

    def _apply_logging_config(self, logging_config: FullLoggingConfig) -> abc.Sequence[logging.Handler]:
        """
        Apply the given logging_config as the current configuration of the logging system.

        This method assume that the given config defines a single root handler.
        """
        handlers_before = list(logging.root.handlers)
        logging_config.apply_config()
        return [handler for handler in logging.root.handlers if handler not in handlers_before]

    @stable_api
    def set_log_level(self, inmanta_log_level: str, cli: bool = True) -> None:
        """
        [DEPRECATED] Set the logging level. A handler should have been created before.
        The possible inmanta log levels and their associated python log level
        are defined in the inmanta.logging.log_levels dictionary.

        :param inmanta_log_level: The inmanta logging level
        :param cli: True if the logs will be outputted to the CLI.
        """
        python_log_level = convert_inmanta_log_level(inmanta_log_level, cli)
        for handler in self._handlers:
            handler.setLevel(python_log_level)
        logging.root.setLevel(python_log_level)

    @stable_api
    def set_log_formatter(self, formatter: logging.Formatter) -> None:
        """
        [DEPRECATED] Set the log formatter. A handler should have been created before

        :param formatter: The log formatter.
        """
        for handler in self._handlers:
            handler.setFormatter(formatter)

    @stable_api
    def set_logfile_location(self, location: str) -> None:
        """
        [DEPRECATED] Set the location of the log file. Be careful that this function will replace the current handler
        with a new one. This means that configurations done on the previous handler will be lost.

        :param location: The location of the log file.
        """
        file_handler = logging.handlers.WatchedFileHandler(filename=location, mode="a+")
        for handler in self._handlers:
            handler.close()
            logging.root.removeHandler(handler)
        self._handlers = [file_handler]
        logging.root.addHandler(file_handler)

    @stable_api
    def get_handler(self) -> logging.Handler:
        """
        [DEPRECATED] Get the logging handler

        :return: The logging handler
        """
        if not self._handlers:
            raise Exception("No handlers found.")
        return self._handlers[0]

    @stable_api
    def register_default_logging_config(self, logging_config: LoggingConfigExtension) -> None:
        """
        Register the default logging config for a certain extension.
        """
        self._logging_configs_extensions.append(logging_config)


class MultiLineFormatter(colorlog.ColoredFormatter):
    """
    Formatter for multi-line log records.

    This class extends the `colorlog.ColoredFormatter` class to provide a custom formatting method for log records that
    span multiple lines.
    """

    inmanta_plugin_pkg_regex = re.compile(r"^inmanta_plugins\.(?P<module_name>[^.]+)")
    # Regex that extracts the name of the module from a fully qualified import of a Python
    # module inside an Inmanta module.

    def __init__(
        self,
        fmt: Optional[str] = None,
        *,
        # keep interface minimal: only include fields we actually use
        log_colors: Optional[LogColors] = None,
        reset: bool = True,
        no_color: bool = False,
        keep_logger_names: bool,
    ):
        """
        Initialize a new `MultiLineFormatter` instance.

        :param fmt: Optional string specifying the log record format.
        :param log_colors: Optional `LogColors` object mapping log level names to color codes.
        :param reset: Boolean indicating whether to reset terminal colors at the end of each log record.
        :param no_color: Boolean indicating whether to disable colors in the output.
        :param keep_logger_names: Display the log messages using the name of the logger that created the log message,
                                  instead of the component of the compiler that was executing while the log record was created
                                  or the name of the module that created the log message.
        """
        super().__init__(fmt, log_colors=log_colors, reset=reset, no_color=no_color)
        self.fmt = fmt
        self._keep_logger_names = keep_logger_names
        self._logger_mode_manager = LoggerModeManager.get_instance()

    def get_header_length(self, record: logging.LogRecord) -> int:
        """
        Get the header length of a given log record.

        :param record: The `logging.LogRecord` object for which to calculate the header length.
        :return: The length of the header in the log record, without color codes.
        """
        # to get the length of the header we want to get the header without the color codes
        formatter = colorlog.ColoredFormatter(
            fmt=self.fmt,
            log_colors=self.log_colors,
            reset=False,
            no_color=True,
        )
        header = formatter.format(
            logging.LogRecord(
                record.name,
                record.levelno,
                record.pathname,
                record.lineno,
                "",
                (),
                None,
            )
        )
        return len(header)

    def format(self, record: logging.LogRecord) -> str:
        """
        Format a log record with added indentation.

        :param record: The `logging.LogRecord` object to format.
        :return: The formatted log record as a string.
        """
        record = self._wrap_record(record)
        indent: str = " " * self.get_header_length(record)
        head, *tail = super().format(record).splitlines(True)
        return head + "".join(indent + line for line in tail)

    def _wrap_record(self, record: logging.LogRecord) -> logging.LogRecord:
        """
        Wrap a log record to perform renaming for specific formatter as determined by the _logger_mode

        This is derived from the way the colorlog.ColoredFormatter works
        """
        old_name = record.name
        new_name = self._get_logger_name_for(old_name)
        if old_name == new_name:
            return record
        attributes = dict(record.__dict__)
        attributes["name"] = new_name
        return logging.makeLogRecord(attributes)

    def _get_logger_name_for(self, logger_name: str) -> str:
        """
        Returns the logger name that should be used in the log record.

        :attr logger_name: The name of the logger that was used to create the log record.
        """
        logger_mode = self._logger_mode_manager.get_logger_mode()
        if not self._keep_logger_names and logger_mode in [LoggerMode.COMPILER, LoggerMode.EXPORTER]:
            if not logger_name.startswith("inmanta"):
                # This is a log record from a third-party library. Don't adjust the logger name.
                return logger_name
            if logger_name == "inmanta.pip":
                # Log record created by a pip subprocess started by the inmanta.
                return "pip"
            match: Optional[re.Match[str]] = self.inmanta_plugin_pkg_regex.match(logger_name)
            if match:
                # Log record created by an Inmanta module.
                return match.groupdict()["module_name"]
            else:
                # Log record created by Inmanta code.
                return logger_mode.value
        else:
            # Don't modify the logger name
            return logger_name


@stable_api
class ParametrizedFileHandler(logging.Handler):
    """
    A file handler that writes to different files based on the last part of the logger name.
    """

    def __init__(self, name_parent_logger: str, log_file_template: str) -> None:
        """
        :param name_parent_logger: The log records created by children of this logger will be written to file by this handler.
                                   This handler will ignore log records created by any other logger.
        :param log_file_template: A template for the path to the log file. This is an f-string that holds the parameter
                                  `child_logger_name`. This part will be replaced by the name of the direct child of
                                  self.name_parent_logger that belongs to the logger that created the record.
        """
        super().__init__()
        self.name_parent_logger = name_parent_logger
        self.log_file_template: str = log_file_template
        self.child_handlers: dict[str, handlers.WatchedFileHandler] = {}

    def flush(self) -> None:
        for child in self.child_handlers.values():
            child.flush()

    def setFormatter(self, fmt: Optional[logging.Formatter]) -> None:
        super().setFormatter(fmt)
        for child in self.child_handlers.values():
            child.setFormatter(fmt)

    def setLevel(self, level: str | int) -> None:
        super().setLevel(level)
        for child in self.child_handlers.values():
            child.setLevel(level)

    def close(self) -> None:
        for child in self.child_handlers.values():
            child.close()
        self.child_handlers = {}
        super().close()

    def emit(self, record: logging.LogRecord) -> None:
        if not record.name.startswith(f"{self.name_parent_logger}."):
            return
        logger_name_without_parent_prefix = record.name.removeprefix(f"{self.name_parent_logger}.")
        logger_name_direct_child_of_parent_logger = logger_name_without_parent_prefix.split(".")[0]
        path_logfile = self.log_file_template.format(child_logger_name=logger_name_direct_child_of_parent_logger)
        if path_logfile not in self.child_handlers:
            try:
                handler = logging.handlers.WatchedFileHandler(filename=path_logfile, mode="a+")
            except FileNotFoundError:
                log_dir = os.path.dirname(self.log_file_template)
                if not os.path.exists(log_dir):
                    LOGGER.warning(
                        "Cannot write to resource action log, because directory %s doesn't exist.",
                        path_logfile,
                    )
                else:
                    LOGGER.exception("Failed to write to resource action log file.")
                return
            handler.setFormatter(self.formatter)
            handler.setLevel(self.level)
            self.child_handlers[path_logfile] = handler
        self.child_handlers[path_logfile].emit(record)


class TornadoDebugLogHandler(logging.Handler):
    """
    A custom log handler for Tornados 'max_clients limit reached' debug logs.
    """

    def __init__(self, level: int = logging.NOTSET) -> None:
        super().__init__(level)
        self.logger = logging.getLogger("inmanta.protocol.endpoints")

    def emit(self, record: logging.LogRecord) -> None:
        if (
            record.levelno == logging.DEBUG
            and record.name.startswith("tornado.general")
            and record.msg.startswith("max_clients limit reached")
        ):
            self.logger.warning(record.msg)  # Log Tornado log as inmanta warnings



"""
    Copyright 2019 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

import datetime
from collections import abc
from enum import Enum
from typing import Optional

from inmanta.stable_api import stable_api


class ResourceState(str, Enum):
    unavailable = "unavailable"  # This state is set by the agent when no handler is available for the resource
    skipped = "skipped"  #
    dry = "dry"
    deployed = "deployed"
    failed = "failed"
    deploying = "deploying"
    available = "available"
    cancelled = "cancelled"  # When a new version is pushed, in progress deploys are cancelled
    undefined = "undefined"  # The state of this resource is unknown at this moment in the orchestration process
    skipped_for_undefined = "skipped_for_undefined"  # This resource depends on an undefined resource


class NonDeployingResourceState(str, Enum):
    unavailable = ResourceState.unavailable.value
    skipped = ResourceState.skipped.value
    dry = ResourceState.dry.value
    deployed = ResourceState.deployed.value
    failed = ResourceState.failed.value
    available = ResourceState.available.value
    cancelled = ResourceState.cancelled.value
    undefined = ResourceState.undefined.value
    skipped_for_undefined = ResourceState.skipped_for_undefined.value


class DeprecatedResourceState(str, Enum):
    """
    Deprecated resource states kept for backwards compatibility.
    """

    # deprecated in iso5
    processing_events = "processing_events"


# undeployable
UNDEPLOYABLE_STATES = [ResourceState.undefined, ResourceState.skipped_for_undefined]
UNDEPLOYABLE_NAMES = [s.name for s in UNDEPLOYABLE_STATES]
# this resource action is not complete, resource is in transient state
TRANSIENT_STATES = [ResourceState.available, ResourceState.deploying]
# not counting as done
NOT_DONE_STATES = TRANSIENT_STATES
# counts as done
DONE_STATES = [
    ResourceState.unavailable,
    ResourceState.skipped,
    ResourceState.deployed,
    ResourceState.failed,
    ResourceState.cancelled,
] + UNDEPLOYABLE_STATES

# starting states
INITIAL_STATES = [ResourceState.available]
# states one can't transition out off
TERMINAL_STATES = UNDEPLOYABLE_STATES
# states on can transition to
VALID_STATES_ON_STATE_UPDATE = [
    ResourceState.unavailable,
    ResourceState.skipped,
    ResourceState.deployed,
    ResourceState.failed,
    ResourceState.deploying,
    ResourceState.cancelled,
    ResourceState.undefined,
    ResourceState.skipped_for_undefined,
]

UNKNOWN_STRING = "<<undefined>>"

"""
States set by server upon upload

1. available
2. undefined (terminal state)
3. skipped_for_undefined (terminal state)

States set by agent
1. skipped
2. failed
3. deployed
4. unavailable
5. cancelled
6. deploying

Each deploy sets the agent state again, all agent states can transition to all agent states

States that are in the action log, but not actual states
1. dry



                                           +-----> deploying        -<--------+
                                           |                                  |
                                           +-----> skipped          -<--------+
                                           |                                  |
                                           +-----> failed           -<--------+
                                           |                                  |
    +---------------->  available  +-------------> unavailable      -<--------+
    |                                      |                                  |
    |                                      +-----> deployed         -<--------+
    |                                      |                                  |
+---+---------+                            +-----> cancelled        -<--------+
| compiler    +------> undefined
+---+---------+
    |
    |
    +----------------> skipped_for_undefined

"""


class Change(str, Enum):
    nochange = "nochange"
    created = "created"
    purged = "purged"
    updated = "updated"


class VersionState(str, Enum):
    success = "success"
    failed = "failed"
    deploying = "deploying"
    pending = "pending"


@stable_api
class ResourceAction(str, Enum):
    """
    Enumeration of all resource actions.
    """

    store = "store"
    push = "push"
    pull = "pull"
    deploy = "deploy"
    dryrun = "dryrun"
    getfact = "getfact"
    other = "other"


STATE_UPDATE = [ResourceAction.deploy]


class AgentTriggerMethod(str, Enum):
    push_incremental_deploy = "push_incremental_deploy"
    push_full_deploy = "push_full_deploy"

    @classmethod
    def get_agent_trigger_method(cls, is_full_deploy: bool) -> "AgentTriggerMethod":
        if is_full_deploy:
            return cls.push_full_deploy
        else:
            return cls.push_incremental_deploy


@stable_api
class LogLevel(str, Enum):
    """
    Log levels used for various parts of the inmanta orchestrator.
    """

    CRITICAL = "CRITICAL"
    ERROR = "ERROR"
    WARNING = "WARNING"
    INFO = "INFO"
    DEBUG = "DEBUG"
    TRACE = "TRACE"

    @property
    def to_int(self) -> int:
        return LOG_LEVEL_AS_INTEGER[self]

    @classmethod
    def _missing_(cls, value: object) -> Optional["LogLevel"]:
        return INTEGER_AS_LOG_LEVEL.get(value, None) if isinstance(value, int) else None


# Mapping each log level to its integer value
LOG_LEVEL_AS_INTEGER: abc.Mapping[LogLevel, int] = {
    LogLevel.CRITICAL: 50,
    LogLevel.ERROR: 40,
    LogLevel.WARNING: 30,
    LogLevel.INFO: 20,
    LogLevel.DEBUG: 10,
    LogLevel.TRACE: 3,
}

INTEGER_AS_LOG_LEVEL: abc.Mapping[int, LogLevel] = {value: log_level for log_level, value in LOG_LEVEL_AS_INTEGER.items()}

INMANTA_URN = "urn:inmanta:"


class Compilestate(str, Enum):
    """
    Compile state, whether the compile did succeed or not
    """

    success = "success"
    failed = "failed"


EXPORT_META_DATA = "export_metadata"
META_DATA_COMPILE_STATE = "inmanta:compile:state"
INMANTA_MT_HEADER = "X-Inmanta-tid"


class ClientType(str, Enum):
    # api: The method with this client type can be used by external clients like the Web-console, cli or 3rd party services
    # compiler: The method with this client type is called by the compiler to communicate with the server
    # agent: The method with this client type is called by the agent to communicate with the server
    api = "api"
    agent = "agent"
    compiler = "compiler"


# For testing

# assume we are running in a tty
ENVIRON_FORCE_TTY = "FORCE_TTY"

LOG_LEVEL_TRACE = 3

NAME_RESOURCE_ACTION_LOGGER = "resource_action_logger"

# Time we give the server/agent to shutdown gracefully, before we force stop the ioloop
SHUTDOWN_GRACE_IOLOOP = 10
# Time we give the server/agent to shutdown gracefully, before we execute sys.exit(3)
SHUTDOWN_GRACE_HARD = 15
# Hard shutdown exit code
EXIT_HARD = 3
# Startup failed exit code
EXIT_START_FAILED = 4

TIME_ISOFMT = "%Y-%m-%dT%H:%M:%S.%f"
TIME_LOGFMT = "%Y-%m-%d %H:%M:%S%z"

PLUGINS_PACKAGE = "inmanta_plugins"

# namespace in which extensions are discovered
EXTENSION_NAMESPACE = "inmanta_ext"
# module inside the extension package that contains the setup function
EXTENSION_MODULE = "extension"

# Default envelope key
ENVELOPE_KEY = "data"

# Max number of attempts when updating modules
MAX_UPDATE_ATTEMPT = 5

# Minimum password length
MIN_PASSWORD_LENGTH = 8


class AgentAction(str, Enum):
    pause = "pause"
    unpause = "unpause"
    keep_paused_on_resume = "keep_paused_on_resume"
    unpause_on_resume = "unpause_on_resume"


class AgentStatus(str, Enum):
    paused = "paused"
    up = "up"
    down = "down"


class ParameterSource(str, Enum):
    fact = "fact"
    plugin = "plugin"
    user = "user"
    report = "report"


class ApiDocsFormat(str, Enum):
    # openapi: the api docs in json format, according to the OpenAPI v3 specification
    # swagger: the api docs in html, using a Swagger-UI view
    openapi = "openapi"
    swagger = "swagger"


class DesiredStateVersionStatus(str, Enum):
    active = "active"
    candidate = "candidate"
    retired = "retired"
    skipped_candidate = "skipped_candidate"


class NotificationSeverity(str, Enum):
    """
    The possible values determine the styling used by the frontend to show them,
    so notifications with severity 'info' will be shown as informational messages,
    the ones with 'error' as error messages and so on.
    The 'message' category corresponds to a generic message, which
    is shown without extra styling on the frontend.
    """

    message = "message"
    info = "info"
    success = "success"
    warning = "warning"
    error = "error"


CF_CACHE_DIR = ".cfcache"

PG_ADVISORY_KEY_PUT_VERSION = 1
PG_ADVISORY_KEY_RELEASE_VERSION = 2
""" lock against releasing a version in an environment, to prevent release races"""


# The filename of the changelog file in an Inmanta module
MODULE_CHANGELOG_FILE = "CHANGELOG.md"


DATETIME_MIN_UTC = datetime.datetime.min.replace(tzinfo=datetime.timezone.utc)

MODULE_PKG_NAME_PREFIX = "inmanta-module-"

# Resource sets marked for deletion during a partial export can be passed via this env
# variable as a space separated list of set ids.
INMANTA_REMOVED_SET_ID = "INMANTA_REMOVED_RESOURCE_SET_ID"
